package Sophos::LogSearch::Constants;

use strict;
use warnings;

use base 'Exporter';

use Date::Parse qw(str2time strptime);
use Time::Local;
use POSIX qw(strftime);
use Sophos::LogSearch::Config qw(%config);
use Sophos::Product::Info;
use Profile;

our $PRODUCT;
our $LOGFILE;


BEGIN {
    $PRODUCT = Sophos::Product::Info->new();
    $LOGFILE = "$Sophos::LogSearch::Config::config{path}->{indexes}/mamelogd.log";

    # Do this at compile time so anything covered by READ_RETRY gets optimized away
    if ($ENV{LOGSEARCH_READ_RETRY}) {
        eval "sub READ_RETRY{1}";
    } else {
        eval "sub READ_RETRY{0}";
    }
};

### Debug constants ###

# For debugging the indexer:
use constant LOGSEARCH_DEBUG => $ENV{'LOGSEARCH_DEBUG'} || 0;
# For debugging the search client:
our $SEARCH_DEBUG = $ENV{'LOGSEARCH_SEARCH_DEBUG'};


our $P = Profile->new(debug => 0); #LOGSEARCH_DEBUG);
use constant CLOSE_STDOUT => 1;


### Data to log ###

# All fields to write to the log files
#our @COLS = qw(time action from msgid relay client to reason subject mseek aliases groups);
our @COLS = qw(time action from msgid relay client to reason subject aliases groups mseek);
our @SLICE_COLS = qw(pfqid aseek zseek);

our @SEARCH_KEYS = qw(from msgid relay to aliases groups mailroute);
our %SEARCH_KEYS = map {$SEARCH_KEYS[$_] => 1} 0..$#SEARCH_KEYS;

our @ALL = (@COLS, @SLICE_COLS);
our %ALL = map {$ALL[$_] => $_} 0..$#ALL;

our $TIME    = $ALL{time};
our $ACTION  = $ALL{action};
our $MSGID   = $ALL{msgid};
our $RELAY   = $ALL{relay};
our $CLIENT  = $ALL{client};
our $TO      = $ALL{to};
our $REASON  = $ALL{reason};
our $SUBJECT = $ALL{subject};
our $MSEEK   = $ALL{mseek};
our $ALIASES = $ALL{aliases};
our $GROUPS   = $ALL{groups};

our $PFQID   = $ALL{pfqid};
our $ASEEK   = $ALL{aseek};
our $ZSEEK   = $ALL{zseek};


### Misc. constants ###
our $DEFER_DAYS = $ENV{DEFER_DAYS} || 4;
our $MAMELOGD_QUEUE_SIZE = $ENV{MAMELOGD_QUEUE_SIZE} || $Sophos::LogSearch::Config::config{queue_max_items} || 1000;
our $MAMELOGD_QUEUE_LIFE = $ENV{MAMELOGD_QUEUE_LIFE} || $Sophos::LogSearch::Config::config{queue_max_time} || 120;
our $CHECK_DISKSPACE_TIME = $ENV{MAMELOGD_CHECK_DISKSPACE_TIME} || 10;
our $PREFIX = $config{prefix};
our $PIDFILE  = $ENV{PIDFILE} || "$PREFIX/var/run/mamelogd.pid";
our $KILL_SIG = 2;
our $CONTINUE = 1;   # Set when we recieve a kill signal

# Maximum number of results to return both:
#     a) as the total number from a search in a cluster
#     b) from a single machine
our $MAX_RESULTS = 1000;

our %EXPORT_TAGS = (
    all => [qw($TIME       $ACTION       $MSGID 
               $RELAY      $TO           $REASON 
               $PFQID      $ASEEK        $ZSEEK 
               $SUBJECT    $MSEEK        $ALIASES 
               $DONE_MAIL  $DONE_MESS    $DONE_BOTH
               @ALL        $GROUPS       $KILL_SIG
               %ALL        %SEARCH_KEYS  @SEARCH_KEYS
               @COLS       $LOGFILE      CLOSE_STDOUT
               $PRODUCT    %INDEXES      LOGSEARCH_DEBUG
               &debug      $DEFER_DAYS   $MAMELOGD_QUEUE_SIZE
               $PREFIX     $PIDFILE      $MAMELOGD_QUEUE_LIFE
               &mame_log   &local_time_string_to_epoch
               LOGSEARCH_DEBUG format_date 
               _set_test_time            &search_debug
               &utc_time_string_to_epoch $P &offset_format
               $CONTINUE   &READ_RETRY   $MAX_RESULTS
	       $CHECK_DISKSPACE_TIME
               &message_is_in_grouplist
	   )],
);

our @EXPORT_OK = @{$EXPORT_TAGS{all}};


### Common subs ###

sub debug {
    return unless LOGSEARCH_DEBUG;
    open my $fv,">>/tmp/mamelog.debug" or die "Can't open debug log";
    print $fv localtime(time) . " ] " . join(' ',@_) . "\n";
    close $fv;
}

my $SEARCH_FH;
sub search_debug {
    return unless ($SEARCH_DEBUG);
    if (!$SEARCH_FH) {
        open $SEARCH_FH, ">$SEARCH_DEBUG" or die "Can't open logsearch debug log '$SEARCH_DEBUG': $!";
    }
    print $SEARCH_FH shift;
}

END { close $SEARCH_FH if $SEARCH_FH };

sub mame_log {
    my ($msg) = @_;

    my $date = strftime('%Y-%m-%dT%H:%M:%S', gmtime);

    open my $log, ">>$LOGFILE" or die "Can't open $LOGFILE for write";
    print $log "$date: $msg\n";
    close $log;
    return;
}


### Date variables and functions ###

my %MONTH_MAP = (
                   'Jan' => 0,
                   'Feb' => 1,
                   'Mar' => 2,
                   'Apr' => 3,
                   'May' => 4,
                   'Jun' => 5,
                   'Jul' => 6,
                   'Aug' => 7,
                   'Sep' => 8,
                   'Oct' => 9,
                   'Nov' => 10,
                   'Dec' => 11,
                );

my $NOW_IS = $ENV{LOG_SEARCH_NOW_IS} || time;
my @NOW_ARRAY = localtime($NOW_IS);

sub _set_test_time {
    $ENV{LOG_SEARCH_NOW_IS} = shift;
    $NOW_IS = $ENV{LOG_SEARCH_NOW_IS} || time;
    @NOW_ARRAY = localtime($NOW_IS);
}

sub _date_to_array {
   my $date = shift;
   # Oct 26 21:44:20  or  Oct 26, 2004 21:44:20
   if ($date =~ /^(\w\w\w) ([\d ]{2})(?:,\s\d{4}\s?)? (\d\d):(\d\d):(\d\d)$/) {
     # $sec,$min,$hours,$mday,$mon,$year
     my @date = ($5,$4,$3,$2,$MONTH_MAP{$1},$NOW_ARRAY[5]); 
     return @date; 
     
   # 2007-08-01T12:49:58 
   } elsif ($date =~ /^(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)$/) {
     # $sec,$min,$hours,$mday,$mon,$year
     my @date = ($6,$5,$4,$3,$2-1,$1 - 1900);
     return @date;

   } else {
      warn "bad date: $date\n";
      return strptime($date);
   }
}

sub local_time_string_to_epoch {
    my ($date) = @_;

    # Minor optimizaton - avoid same second lookups
    my $now = $ENV{LOG_SEARCH_NOW_IS} || time;
    if ($now != $NOW_IS) {
        $NOW_IS    = $now;
        @NOW_ARRAY = localtime($NOW_IS);
    }
 
    $P->begin('local_time_string_to_epoch:strptime');
    my @parsed =  _date_to_array($date);
    if (scalar(@parsed) != 6 or grep { not defined $_ } @parsed) {
        warn "invalid date: $date - " . join(',',@parsed) . "\n"; 
        return;     
    }
    $P->end('local_time_string_to_epoch:strptime');

    # Use this/now year.
    $parsed[5] = $NOW_ARRAY[5] unless $parsed[5];

    $P->begin('local_time_string_to_epoch:time_local');
    my $epoch = eval { timelocal(@parsed) };
    if ($@ or !$epoch) { 
        $P->end('local_time_string_to_epoch:time_local');
        return;
    }

    # if the time is in the future, it is in the past, use last
    # year as the date
    if ($epoch > $NOW_IS) {
        $parsed[5] = $NOW_ARRAY[5] - 1;
        $epoch = timelocal(@parsed);
    }

    $P->end('local_time_string_to_epoch:time_local');
    return $epoch;
}

sub utc_time_string_to_epoch {
  my $time = shift;
  $P->begin('utc_time_string_to_epoch');
  my @now = _date_to_array($time);
  my $val = timegm(@now); 
  $P->end('utc_time_string_to_epoch');
  return $val;
}

sub format_date {
  my ($sec,$min,$hour,$mday,$mon,$year) = @_;
  $P->begin("strftime");
  # sec, min, hour, mday, mon, year, wday
  my $v = sprintf('%04d-%02d-%02dT%02d:%02d:%02d',$year + 1900, $mon + 1, $mday, $hour,$min,$sec); 
  $P->end("strftime");
  return $v;
}

sub offset_format {
   my $int = shift;
   my $len = length($int);
   die "Integer too large" if $len > 10;

   my $pad = '0' x (10 - $len);
   return "${pad}${int}";
}

### Other common functions ###

sub message_is_in_grouplist {
    my ($msg, $groups) = @_;

    if($groups && @$groups){
        my $msg_groups = [ split /,/, $msg->field('groups') ];
        
        my $group_matches = 0;
        for my $msg_grp (@$msg_groups){
            for my $grp (@$groups){
                if($grp eq $msg_grp){
                    $group_matches = 1;
                    last;
                }
            }
        }

        return undef if !$group_matches;
    }

    return $msg;
}

1;
