package Sophos::LogSearch::Local;

use strict;
use warnings;

use Carp qw(croak);
use POSIX qw(strftime);
use Cwd qw(cwd);
use Net::Domain qw(hostname);
use IPC::Open2 qw(open2);

use Sophos::LogSearch::Message;
use Sophos::LogSearch::Config;
use Sophos::LogSearch::Constants qw(:all);
use Sophos::LogSearch::Source::Indexer;

my $cwd;

sub new {
    my ($class) = @_;
    my $self = {
        hostname => hostname(),
        message_limit => $MAX_RESULTS,
    };
    bless $self, $class;
}

sub search {
    my ($self, %args) = @_;

    my $offset = $args{offset};
    my $count = $args{count};

    croak "LogSearch::search(): 'offset' is required" unless defined $offset;
    croak "LogSearch::search(): 'count' is required" unless defined $count;
    
    $cwd = cwd;
    chdir $Sophos::LogSearch::Config::config{path}->{indexes} or die "Couldn't chdir $Sophos::LogSearch::Config::config{path}->{indexes}: $!";
    END { chdir $cwd if $cwd };
    
    # unmarshal groups information - required when search is called via 'remote'.
    if ($args{groups} && !ref($args{groups})) {
        $args{groups} = [ split(',', $args{groups}) ];
    }

    my $cmds = $self->_commands(%args); 
    
    my $slicer_cmd = $$cmds[0];
    my $filter_cmd = $$cmds[1];
    my $sort_cmd   = $$cmds[2];
    
    # The slicer commands can be massive (i.e. 'n' bytes > getconf 'ARG_MAX')
    # To avoid breaching the shell limit, the slicer commands are written to
    # a read/write pipe i.e. write 'index info' | sh | filter | sort.    

    my $cmd = "/bin/sh ";
    $cmd .= $filter_cmd if ($filter_cmd);
    $cmd .= $sort_cmd if ($sort_cmd); 
    
    my @msgs;
    my ($pid);
    eval {
        pipe(my $PARENT_READ, my $CHILD_WRITE) or die "Could not create pipe: $!";
        pipe(my $CHILD_READ, my $PARENT_WRITE) or die "Could not create pipe: $!";

        $pid = fork();
        die "Could not fork: $!" if (not defined $pid);

        #child
        if ($pid == 0) {
            close($PARENT_READ); close($PARENT_WRITE);
            POSIX::dup2(fileno($CHILD_READ), 0);
            POSIX::dup2(fileno($CHILD_WRITE), 1);

            exec $cmd;
        }
        #parent
        else {
            close($CHILD_READ); close($CHILD_WRITE);

            print $PARENT_WRITE "$slicer_cmd\n";

            close($PARENT_WRITE);

	    my $m = Sophos::LogSearch::Message->new();
            while (my $cnt++ <= $self->{message_limit} and my $line = <$PARENT_READ>) {
                # Code at bottom does the offset...
                # only store data after $offset and before $offset + $count

                push @msgs, $m->line(mline => $line);
                $msgs[-1]->{host} = $self->{hostname};
            }
            close($PARENT_READ) or die "$! : Failed to close reader\n";
            waitpid $pid, 0; # avoid zombie child
        }
    };

    unless ($@) {
        my $sort = $args{sort} || '-time';
        my $desc = ($sort =~ s/^-//) ? -1 : 1;
        @msgs = sort { $desc * ($a->{$sort} cmp $b->{$sort}) } @msgs;

        # Ensure we only return MAX results.
        $count = ($count < $self->{message_limit}) ? $count : $self->{message_limit};
        @msgs = splice(@msgs, $offset, $count);
    }

    chdir $cwd or die "Can't chdir $cwd: $!";
    if ($@) {
        warn $@ . "\n";
        return [];
    }

    return \@msgs;
}


sub message {
    my ($self, $key, $groups) = @_;
    my ($host,$date,$offset) = split /\~/, $key;
    my $slicer = $PRODUCT->get_util('slicer');
   
    return unless $host and $date and $offset;

    my $mline = qx(echo $offset | $slicer $Sophos::LogSearch::Config::config{path}->{indexes}/$date/mamelog);
    chomp $mline; # PROBABLY a good idea. possibly redundant
    return unless $mline;

    my $msg = Sophos::LogSearch::Message->new->line(mline => $mline);
    return message_is_in_grouplist($msg,$groups);
}


sub _commands {
    my ($self, %search_filter) = @_;
    my $gunzip_to_stdout = $PRODUCT->get_util('gunzip_to_stdout');
    my $egrep = $PRODUCT->get_util('egrep');
    my $join = $PRODUCT->get_util('join');
    my $sort = $PRODUCT->get_util('sort');
    my $head_n = $PRODUCT->get_util('head_n');
    my $mkfifo = $PRODUCT->get_util('mkfifo');
    my $slicer = $PRODUCT->get_util('slicer');

    use constant IX_DATE => 0;
    use constant IX_START_OFFSET => 1;
    use constant IX_END_OFFSET => 2;

    my $sort_order = (delete $search_filter{sort}) || '-time';
    my $atime = delete $search_filter{atime} || ''; # Non-numeric less than any yyyy-mm-dd.
    my $ztime = delete $search_filter{ztime} || ''; # Non-numeric greater than any yyyy-mm-dd.

    # find all our dirs (with a date format name) from the log directory.
    # Only pick the ones that are within time range if we have one
    my @dirs = grep {(!$atime or $_ ge substr($atime,0,10)) and 
             (!$ztime or $_ le substr($ztime,0,10))
            } glob "20??-??-??";

    @dirs = reverse @dirs unless $sort_order eq 'time';
    $atime ||= '-';
    $ztime ||= '-';

    my $S = '[^ ]*';        # "egrep" does not understand '\S*'

    my %fifos;              # Any indexed fields need (gunzip>keyfifo&) commands?
    my @cmds;

    for my $dir (@dirs) {
        my $cmd = "";
        my $base = "$dir/mamelog"; 

        for my $key (grep {$search_filter{$_}} @SEARCH_KEYS) {

            if (ref($search_filter{$key}) eq 'ARRAY') {
                $search_filter{$key} = join('|', @{$search_filter{$key}});
            }

            my $file = "$base.$key";
            if ($cmd) {
                if (-f "$file.gz") { 
                    $fifos{$key} = 1;
                    $cmd = "$gunzip_to_stdout ${file}.gz >/tmp/$key.$$&\n$cmd"; 
                    $file = "/tmp/$key.$$"
                }
                $cmd .= "| $join 2.1 2.2 - $file | $egrep '$search_filter{$key}$S\$' -";
            } else {
                if (-f "$file.gz") {
                    # it's compressed already, grep the output of uncompressing it
                    $cmd = "$gunzip_to_stdout $file.gz | $egrep '$search_filter{$key}$S\$' - ";
                } else {
                    # otherwise just grep it
                    $cmd = "$egrep '$search_filter{$key}$S\$' '$file'";
                }
            }
        }

        my @offsets;
        my %ranges;
        if (-f "$dir/mamelog.time") {
            open my $fh, "$dir/mamelog.time" or die "cannot open time index: $!";
            #open my $fh, "$dir/mamelog.time" or (system 'zsh' and die "cannot open time index: $!");
            while (<$fh>) {
                my @t_ix = split ' ';
                $ranges{$t_ix[IX_DATE]} = [$t_ix[IX_START_OFFSET],$t_ix[IX_END_OFFSET]];
            }
            close $fh;
        }

        # Pick the files that are relevant for our date range
        my $last = "999999999"; # MAGIC BIG NUMBER

        # NumZeros
        my @all_files = glob("$dir/mamelog.?????????[0-9]{,.gz}");
        my %gzfiles = map { $_ => 1 } grep { /\.gz$/ } @all_files; # only files that end in gz 
        my @files;
 
        foreach my $file ( grep { /[0-9]$/ } @all_files ) { # only non .gz files
           delete($gzfiles{"$file.gz"}) if exists $gzfiles{"$file.gz"};
           $gzfiles{$file} = 1;
        }
        @files = keys %gzfiles;

        for my $file (sort { $b cmp $a } @files  ) {

                                 # NumZeros
            my ($off) = $file =~ /mamelog\.(\d{10})/;

            if (my $range = $ranges{$off}) {
                if (($atime eq '-' or $atime le $range->[1]) and 
                    ($ztime eq '-' or $ztime ge $range->[0])) {
                    push @offsets, "\t$off\t$last";
                }
            }
            else {
                push @offsets, "\t$off\t$last";
            }
            
            $last = $off - 1;
        }

        # we always want to look backwards through segments so that head -1000 
        # works properly unless we are sorting by time ascending
        @offsets = reverse @offsets if $sort_order eq 'time';

        if (@offsets) {    # Empty if no mamelog{,.time} for today
            my $offsets = join "\n", @offsets;

            $cmd .= $cmd ? "| $slicer $base $atime $ztime" : 
                   "echo '$offsets' | $slicer $base $atime $ztime";
               
            push @cmds, $cmd;
        }
    }

    my $cmds = join("\n", @cmds);

    # turn all the gunzip_to_stdout output files in /tmp to fifo's so that our
    # huge pipeline of commands works
    
    if (my $fifos = join " ", map {"/tmp/$_.$$"} keys %fifos) {
        $cmds = "$mkfifo $fifos\n$cmds\nrm -f $fifos\n";
    }

    my $max = $self->{message_limit};

    # if there is a search_filter key that isn't a key we recognize as a
    # searching key then filter by it now using grep, ie. postprocess our result
    # after we have constructed our cmd pipeline
    
    # We need to search for all messages with 'reject', 'bounced', or 'deferred' action if the
    # search action is 'reject'.
    $search_filter{action} = '(Bounced|Deferred|Reject)' 
        if $search_filter{action} && $search_filter{action} eq 'Reject';
    my $patt = join " ", map {$search_filter{$_} && !$SEARCH_KEYS{$_}?  "${S}$search_filter{$_}${S}" : $S} @ALL;
    # Remove all $S's ([^ ]*) from the end of the pattern
    # [^ ]* [^ ]*Bounced[^ ]* [^ ]* [^ ]* [^ ]* => [^ ]* [^ ]*Bounced
    my $regex = qr/\Q$S\E/;
    $patt =~ s/(\s?$regex)*$//;
    $patt = "| $egrep '^$patt'" if $patt;
    $patt .= "|$head_n" . $max;

    my @sort_by = split(/,/,$sort_order);
    my @sort_cmd_line;

    my $idx = Sophos::LogSearch::Source::Indexer->new();

    foreach my $sort_by_key (@sort_by) {
      
      my $cmd_line = '';

      if ($sort_by_key =~ /^(-)(.*)/) {
          $sort_by_key = $idx->column_number_of_field(uc $2);
          $cmd_line = "-k " . ($sort_by_key + 1) . " -r";
      } else {
          $sort_by_key = $idx->column_number_of_field(uc $sort_by_key);
          $cmd_line = "-k " . ($sort_by_key + 1); 
      }

      push(@sort_cmd_line, $cmd_line); 
    }
    
    $sort_order = "| $sort " . join(' ',@sort_cmd_line);

    my @commands;
    if ($cmds) {
        @commands = ($cmds, $patt, $sort_order);
    } else {
        push @commands, "cat /dev/null\n";
    }

    return \@commands;
}

1;

