package Sophos::LogSearch::MTA::Postfix;

use strict;
use warnings;

use base 'Exporter';
our @EXPORT_OK = qw/mame_log/;

use Time::HiRes                  qw(time);

use Sophos::LogSearch::Constants qw/LOGSEARCH_DEBUG &debug &mame_log $PRODUCT/;
use Sophos::LogSearch::Config qw(%config);
use Sophos::LogSearch::Transaction;
use PureMessage::Group::CDBMap;
use PureMessage::Map;

my %reasons = (
    virus => 'virus',
    spam => 'spam',
    keyword => 'keyword',
    suspect => 'suspect',
    encrypted => 'encrypted',
    cantscan => 'unscannable',
    offensive => 'offensive',
    legit => 'legitimate',
);

my %actions = (
    quarantine => 'quarantined',
    deliver => 'delivered',
    discard => 'discarded',
    reject => 'rejected',
    route => 'routed',
);


# ===============
# Regexes

# Mar 30 14:58:13 vedder postfix/smtpd[13036]: NOQUEUE: reject: RCPT from tang.ActiveState.com[192.168.3.209]: 565 5.7.1 <tang.ActiveState.com[192.168.3.209]>: Client host rejected: keep buggyd outta my system!; from=<sender5blocked@stdinject5.sophoesexample.com> to=<pmx@vedder.ca.sophos.com> proto=ESMTP helo=<localhost.localdomain>

# solaris
# Jun 21 13:14:34 acheron postfix/smtpd[17716]: [ID 197553 mail.info] NOQUEUE: reject: RCPT from tang.ActiveState.com[192.168.3.209]: 550 5.1.1 <pmxtgr3a@acheron.ca.sophos.com>: Recipient address rejected: User unknown in local recipient table; from=<sender7blocked@stdinject7.sophoesexample.com> to=<pmxtgr3a@acheron.ca.sophos.com> proto=ESMTP helo=<localhost.localdomain>
# postfix/smtpd[634]: [ID 197553 mail.info] NOQUEUE:

my $reject_line_re = qr{postfix[^:]+:
             (?:\s+\[[^\]]+\])?   # possibly eat the solaris syslog trash
		     \s+([^:]+):		# Postfix queue id || NOQUEUE
		     \s+reject:
		     \s+\S+				# What is rejected
		     \s+from\s+([^[]+\[[^]]+\]):	# Relay
		     \s+(\d+)				# Return code
             (?:\s+[\d\.]+)*    # Optional extended status code
		     \s+<[^>]+>:		# Denied addr
		     \s+[^;]+;			# Error Message
		     (?:
		     \s+from=<([^>]+)>	# From
		     \s+to=<([^>]+)>	# To
		     )?
		    }xsm;

my $rejectable_re = qr{(?:
			   queue\sfile\ssize\slimit\sexceeded
		       )}xsm;

my $join_line_re = qr{, sent ([0-9_A-F]+) ?([0-9A-F]*)};


sub new {
    my $class = shift;
    my $map;
    eval {
        $map = PureMessage::Map->new('recipient-aliases');
    };
    if (!$map or $@) {
        mame_log("Error initializing recipient-aliases map: $@");
        die "Unable to initialize recipient-aliases map.";
    }
    return bless { recipient_aliases => $map }, $class;
}


sub aliases {
    my ($self,$to) = @_;
    $self->{recipient_aliases}->refresh() if $self->{recipient_aliases}->can('refresh'); 
    return $self->{recipient_aliases}->transform($to);
}


# Process a Postfix log line
sub process_line {
    my ($self, $line) = @_;

    my %mail_info;

    debug("MTA: process_line");
    unless (exists $self->{gmap}) {
        $self->{gmap} = PureMessage::Group::CDBMap->Tie;
    }
    my $gmap = $self->{gmap};

    # Blocked message lines
    if (my ($pfqid, $relay, $rcode, $from, $to) = $line =~ $reject_line_re) {
        debug("MTA: line is a reject line");
        
        my ($relay_ip) = $relay =~ /\[([^]]+)\]/;

        $from   ||= '-';
        $to     ||= '-';

        # 'NOQUEUE' lines have no postfix queue id so create a fake one
        
        my $pfid = $pfqid eq 'NOQUEUE' ? 'NOQUEUE_' . Time::HiRes::time() : $pfqid;

        my @aliases = $self->aliases($to);
        my @groups;
        foreach my $a (@aliases) {
               # XXX: Should we be checking for duplicates?
               my $group = $gmap->lookup($a);
               push @groups, $group if defined($group);
        }
        my $to_group = $gmap->lookup($to);
        push @groups, $to_group if $to_group;

        my $status = $rcode =~ /^4/ ? 'deferred' : 'bounced';
        %mail_info = (
            id => $pfid,
            done_mail_log => 1,
            done_message_log => 1,  # We'll never see the message log for a blocked message,
                                    # so mark it as done anyway
            STATUS => {$pfid => {$to => $status}},
            CLIENT => $relay_ip,
            FROM => $from,
            TO => {$to => 1},
            RELAY => $relay_ip,
            MAILROUTE => $relay,
            ALIASES => join(',', @aliases),
            GROUPS => join(',', @groups),
        );

# Jun 25 17:17:47 acheron postfix/smtpd[10539]: [ID 197553 mail.info] 711CA6C5C2: client=tube.ca.sophos.com[192.168.99.60]
    } elsif (my ($pfid) = $line =~ / postfix[^ ]*:(?:\s+\[[^\]]+\]\s+)?\s*[a-z: ]*([0-9A-F]+):/) {
        debug("MTA: $pfid: postfix line found: '$line'");

            %mail_info = (
                id => $pfid,
            );

            # If it's a reject line, mark it as done
            if ($line =~ $rejectable_re) {
                $mail_info{done_mail_log} = 1;
                $mail_info{done_message_log} = 1;
                $mail_info{ACTION} = 'bounced';
                debug("MTA: $pfid: reject line: ACTION='bounced'");
            }

            # postfix/backend/pickup[94963]: 261751505F98: uid=2001 from=<kevinj@ca.sophos.com> orig_id=D7E541505F89
            if ($line =~ / orig_id=(\S+)/) {
                $mail_info{ORIG_ID} = $1;
                debug("MTA: $pfid: ORID_ID='$1'");
            } elsif ($line =~ / from=<([^>]+)>/) {
                # postfix/backend/qmgr[68959]: 261751505F98: from=<kevinj@ca.sophos.com>, size=1240, nrcpt=1 (queue active)
                $mail_info{FROM} = $1;
                debug("MTA: $pfid: FROM='$1'");

            } elsif (my ($client) = $line =~ / client=([^\s]+)/) {
                my ($fqdn, $ip) = $client =~ m{^([^[]+)\[(.+)\]};
                $client =~ s/ /_/g;
                if (_fingerprint_host($ip)) {
                    $mail_info{CLIENT} = $ip;
                    push @{$mail_info{MAILROUTE}}, $client;
                }
                debug("MTA: $pfid: CLIENT='$mail_info{CLIENT}'");
                debug("MTA: $pfid: but we ignored the client") if $client eq $mail_info{CLIENT};

            } elsif (my ($to,$relay,$status) = $line =~ / to=<([^>]+)>,.*relay=([^,]+),.*status=([^ ]+)/) {

                debug("MTA: $pfid: delivery line: TO='$to', RELAY='$relay', STATUS='$status'");
                $mail_info{STATUS}{$pfid}{$to} = $status;
                $mail_info{TO}{$to} = 1;

                my $ip = $relay;
                $ip =~ s/^[^[]+\[(.*)\](?::\d+)?$/$1/;
                if (_fingerprint_host($ip)) {
                    $mail_info{RELAY} = $ip;
                    push @{$mail_info{MAILROUTE}}, $relay;
                }

                # 046A244093A: to=<root@localhost.ca.sophos.com>, orig_to=<root@localhost>, relay=127.0.0.1[127.0.0.1]:10025, delay=1.4, delays=1.3/0.02/0/0.12, dsn=2.0.0, status=sent (250 OK, sent 460AB773_22542_11_1 4C5B9440941) 
                if ($line =~ $join_line_re) {
                    my $milter_id = $mail_info{milter_id} = $1;
                    my $otherid = $2;
                    if ($otherid) {
                        $mail_info{id2} = $otherid;
                    }

                    debug("MTA: $pfid: sent line: MILTER_ID='$milter_id', ID2='$otherid'");
                    if ($line =~ /relay=([^,]+)/) {
                        #$mail->{MILTER_HOST} = $1;
                        # Use relay line at this point to grab the milter host
                        push @{$mail_info{MAILROUTE}}, $1 if (!exists($mail_info{MAILROUTE}));

                        # Remove relay since it's the milter host
                        delete($mail_info{RELAY});
                    }
                }

                if ($status =~ /^(?:deferred|bounced|undeliverable)$/) {
                    debug("MTA: $pfid mail deferred/bounced");
                    $mail_info{done_mail_log} = 1;
                    if ($ip =~ /\Q127.0.0.\E/ || $status eq 'undeliverable') {
                        $mail_info{CLIENT} = $ip;
                        # We are leaving an orphaned mail_log transaction behind!!!
                        $mail_info{done_message_log} = 1;
                    }
                }

            } elsif ($line =~ / message-id=<([^>]*)>/) {
                debug("MTA: $pfid: MSGID='$1'");
                $mail_info{MSGID} = $1;

            } elsif ($line =~ / removed$/) {
                # This should be the last line we ever see for this ID
                debug("MTA: $pfid: removed line found");
                $mail_info{done_mail_log} = 1;
                debug("MTA: $pfid is done mail.log");
            }

            if (LOGSEARCH_DEBUG) {
                eval "require Data::Dumper;";
                "Data::Dumper"->import();
                debug("MTA: mail: " . Dumper(\%mail_info)); 
            }
#Apr  7 14:27:14 dolly queue[5950]: 8910C3C40FD: relay=eko.ca.sophos.com:25 status=sent (250 2.0.0 Ok: queued as 928024674D7)
    } elsif ($line =~ / queue\[\d+\]:\s+([0-9A-Z_]+):(.*)\srelay=(\S+)\sstatus=(\w+)/) {
        my ($pfid, $other, $relay, $status) = ($1, $2, $3, $4);
        debug("MTA: $pfid: queue line found: '$line'");
        %mail_info = (
            id => $pfid,
            RELAY => $relay,
            done_mail_log => 1,            
        );
        
        if ($other =~ /from=(\S+)\s+to=(\S+)\s+client=(\S+)\s+subject=(\S+)\s+msgid=(\S+)/) {
            my ($from, $client, $subject, $msgid) = ($1, $3, $4, $5);
            my @to = split ',', $2;
            $mail_info{TO} = {map { $_, 1 } @to};
            $mail_info{STATUS}{$pfid} = {map { $_ => $status } @to};
            $mail_info{ACTION} = join ',', map {'route'} @to;
            $mail_info{FROM} = $from;
            $mail_info{CLIENT} = $client;
            $mail_info{SUBJECT} = $subject;
            $mail_info{MSGID} = $msgid;
            $mail_info{route_copy} = 1;
         
            my %groups;
            my @aliases;
            for my $addr (@to) {
                # Group for this to address
                my $to_group = $gmap->lookup($addr);
                $groups{$to_group} = 1 if defined $to_group;
                # Aliases for this to address
                my @aliases_for_addr = $self->aliases($addr);
                push @aliases, @aliases_for_addr;
                # Groups for each aliases
                for my $a (@aliases_for_addr) {
                    my $group = $gmap->lookup($a);
                    $groups{$group} = 1 if defined $group;
                }
            }
            $mail_info{ALIASES} = join ',', @aliases;
            $mail_info{GROUPS}  = join ',', keys %groups;
        }
    } else {
      #warn "unknown messagelog format: $line\n";
    }

    if (exists $mail_info{MAILROUTE} && (ref($mail_info{MAILROUTE}) eq 'ARRAY') ) {
        $mail_info{MAILROUTE} = join(',', @{$mail_info{MAILROUTE}});
        debug($mail_info{id} . ": MAILROUTE='$mail_info{MAILROUTE}'");
    }

    if (keys %mail_info) {
        return Sophos::LogSearch::Transaction->new(%mail_info);

    } else {
        return undef;
    }
}


# Should we record information about this IP address?
sub _fingerprint_host {
    my $addr = shift;
    return 0 if (!$addr || $addr eq 'none');
    return 1 if ($Sophos::LogSearch::Config::config{index_local_delivery});
    return 1 if ($addr ne 'local'
                 and $addr !~ '^127.0.0'
                 and $addr !~ '^localhost[^a-z]?');
    return 0;
}


# ===============
# Returning and parsing log lines

sub logline_data {
    my ($self, %args) = @_;
    my $rawlines = $args{rawlines};
    my @actions = split(',', delete $args{action});
    my @reasons = split(',', delete $args{reason});

    my $logline = qr{ (\d+-+\d+-\d+\s+\d+:\d+:\d+)\s # date
                      \S+\s                          # hostname
                      (.*?)\[\d+\]:\s*               # process
                      ([a-z: ]*)                     # optional log prefix (e.g. "warning:") 
                      [^:]+:\s		                 # qid
                      (.+)			                 # log
                  }xsm;

    my $queue = 'unknown';

    my %re = (
        'postfix/smtpd' => [
            qr{^client= [^[]+ \[ ( [^]]+ ) \]}xsm =>
                sub {[Received => "Connection from $_[1]"]},

            qr{^reject: [^:]+: [^:]+: Client host rejected} =>
                sub {[Rejected => "For reason: Message source blocked by IP Reputation."]},

            qr{^reject: .* host \S+ said: ([^(;]+)} =>
                sub {[Rejected => ucfirst $_[1]]},

            qr{^reject: [^:]+: [^<]+<([^>]+)>: ([^;]+)} =>
                sub {[Rejected => "$_[1]: $_[2]"]},

            qr{^reject: [^:]+: [^:]+: ([^;]+)} =>
                sub {[Rejected => ucfirst $_[1]]},
        ],

        'postfix/qmgr' => [
            qr{relay=127\.0\.0\.1\[127\.0\.0\.1\], [^,]+, status=sent} =>
                sub {[Scanned => "At $_[1]"]},
        ],

        'postfix/smtp' => [
            qr{status=undeliverable \((.*)\)} =>
            sub {[Undeliverable => $_[1]]},

            qr{relay=[^[]+\[([^]]+)\].*\(250 ([^)]+)\)} => 
            sub {
                my $next_reason = shift(@reasons) || 'legit';
                $next_reason = 'legit' if $next_reason eq '-';
                my $reason = $reasons{lc($next_reason)} || $next_reason;
                my $next_action = shift(@actions) || '-';
                my $action = $actions{lc($next_action)} || '?';
                [Scanned => "With result: $reason. Message will be $action."];
            },
        ],

        'postfix/cleanup' => [
            qr{^warning: ([^(]+)} =>
            sub {$_[1] =~ /queue file size limit exceeded/ ? 
                    [Rejected => ucfirst $_[1]] : 
                    [Warning => ucfirst $_[1]]
            },
        ],

        'postfix/qmgr' => [
            qr{^removed$} =>
            sub {[Removed => "From the pre-filter $queue queue at $_[0]"]},

            qr{\(queue ([^)]+)\)$} => 
            sub {$queue = $_[1]; [Queued => "In the pre-filter $_[1] queue at $_[0]"]},
        ],

        'postfix/backend/qmgr' => [
            qr{^removed$} =>
            sub {[Removed => "From the delivery $queue queue at $_[0]"]},

            qr{\(queue ([^)]+)\)$} => 
            sub {$queue = $_[1]; [Queued => "In the delivery $_[1] queue at $_[0]"]},
        ],

        'milter' => [
            qr{^POLICY WARNING: ([^(]+)} =>
            sub {[Warning => ucfirst $_[1]]},
        ],

        'postfix/backend/smtp' => [
            qr{status=deferred \(([^)]+)\)} => 
            sub {[Deferred => ucfirst "$_[1] at $_[0]"]},

            qr{relay=[^[]+\[([^]]+)\].*\(250 ([^)]+)\)} => 
            sub {[Delivered => "To $_[1] at $_[0] with response '$_[2]'"]},
        ],
        
        'queue' => [
            qr{relay=(.+)\s+status=sent} =>
            sub {[Routed => "to $_[1] at $_[0]. Status sent."]},
            
            qr{relay=(.+)\s+status=rejected\s+\(.+:\s+(.+)\s+at} =>
            sub {[Routed => "to $_[1] at $_[0]. Status rejected, $_[2]"]},
            
            qr{relay=(.+)\s+status=deferred\s+\((.+)\)} =>
            sub {[Routed => "to $_[1] at $_[0]. Status deferred, $_[2]"]},
        ],

    );

    my @data;

    # Iterate over the regexps and run the sub above passing the date plus the
    # list of matched patterns

    for my $line (split "\n", $rawlines) {
        my ($date, $proc, $prefix, $log) = $line =~ $logline or next;
        $log = "$prefix$log";
        
        my $proc_re = $re{$proc};
        for my $i (0 .. $#$proc_re) {
            next if $i & 1; #skip the odds
            my $re = $proc_re->[$i]; 
            my $sub = $proc_re->[$i+1];
            my @vals = $log =~ $re;
            next unless @vals;
            push @data, my @res = $sub->($date, @vals);
            # warn "matched $re: " . join(',', map { @$_ } @res) . "\n";
            last;
        }
    }

    my $action = shift @actions || '-';

    unless ($action =~ /^(?:reject|bounced|deferred|deliver)$/i) {
        my $final_action = ucfirst $actions{lc($action)} || $action;
        my $reason = shift @reasons || 'legit';
        $reason = 'legit' if $reason eq '-';

        if ($final_action and $final_action ne '-') {
            my $final_reason = $reasons{$reason} ? $reasons{$reason} : $reason;
            push @data, [$final_action => "For reason: $final_reason"];
        }
    }

    return \@data;
}



sub get_milter_id {
    my ($self, $loglines) = @_;

    foreach (split /\n/, $loglines) {
        if (my ($milter_id) = $_ =~ m{postfix/smtp.+relay.+ sent\s([\w_]+)\s}) {
           return $milter_id; 
        }
    }

    return;
}


# Divide a string of loglines into incoming and outgoing loglines
sub get_divided_loglines {
    my ($self, $loglines) = @_;

    my ($incoming, $outgoing) = ('', '');

    my $postfix_id_re = qr{(queue|postfix)[^:]+:\s+([^:]+):};

    my $incoming_id;
    foreach my $line (split /\n/, $loglines) {

        next unless $line =~ $postfix_id_re;
        my ($proc, $pfid) = ($1, $2);

        $incoming_id = $pfid if !$incoming_id;

        if (my ($milter_id, $id2) = $line =~ $join_line_re ) {

            # If we think that ID2 is the incoming ID, switch the two
            if ($id2 && ($incoming_id eq $id2)) {

                    # Switch the IDs
                    $incoming_id = $pfid;

                    # Switch the loglines
                    my $temp = $incoming;
                    $incoming = $outgoing;
                    $outgoing = $temp;
            }
        }

        if ($proc eq 'queue') {
            $outgoing .= "$line\n";
        }
        elsif ($pfid eq $incoming_id) {

            $incoming .= "$line\n";

        } else {

            $outgoing .= "$line\n";

        }
    }

    return ($incoming, $outgoing);
}


sub get_milter_host {
    my ($self, $loglines) = @_;

    foreach (split /\n/, $loglines) {
        if (my ($host_ip) = $_ =~ m{postfix/smtp.+relay=([^\[]+).+ OK, sent\s[\w_]+\s}) {
            return $host_ip; 
        }
    }

    return;
}

1;
