package Sophos::LogSearch::Message;

use strict;
use warnings;
use Sophos::LogSearch::Constants qw($PRODUCT);
use Sophos::LogSearch::MTA::Postfix;

sub new {
    my ($class, %args) = @_;
    my $self = bless \%args, $class;

    $self->{slicer} = $PRODUCT->get_util('slicer');
    $self->{mta} = Sophos::LogSearch::MTA::Postfix->new();
    $self->{host} ||= 'localhost';

    return $self;
}

sub line {
    my $self = shift;
    my %args = map { $_ => $self->{$_} } keys %$self;
    return Sophos::LogSearch::Message::Content->new(%args, @_);
}

1;

package Sophos::LogSearch::Message::Content;

use strict;
use warnings;
use POSIX qw(strftime);
use Sophos::LogSearch::Constants qw(@COLS $PRODUCT 
                                    local_time_string_to_epoch &search_debug);
use Sophos::LogSearch::Config;

use Carp qw(croak);

sub new {
    my ($class, %args) = @_;
    my $self = bless \%args, $class;

    if (my $mline = delete $args{mline}) {
        $self->_parse_mline($mline);
        $self->{id} = [substr($self->{time},0,10), $self->{mseek}];
    }
    elsif (my $id = $args{id}) {
        $self->_parse_mline($self->_get_mline(@{$args{id}}));
    }

    $self->{fetchkey} = join('~', ($self->{host}, substr($self->{time},0,10), $self->{mseek}) );

    return $self;
}

sub no_log_data {
    my $self = shift;
    $self->{_no_log_data} = 1;
}

sub _get_mline {
    my ($self,$dir,$mseek) = @_;
    return qx{echo $mseek | $self->{slicer} $Sophos::LogSearch::Config::config{path}->{indexes}/$dir/mamelog};
}

sub _parse_mline {
    my ($self, $mline) = @_;

    search_debug("mline: '$mline'");
    $self->{mline} = $mline;

    my @values = split ' ', $mline;
    $self->{$COLS[$_]} = $values[$_] for 0 .. $#COLS;

    $self->{relay} =~ s/!+$//;

    my @relays = split(',', $self->{relay});
    my @formatted_relays = ();

    foreach my $r (@relays) {
        if ($r =~ s/^route=//) {
            $self->{routes} = [split('~', $r)];
        }
        elsif ($r =~ s/^copy=//) {
            $self->{copies} = [split('~', $r)];
        }
        push @formatted_relays, $r;
    }
    $self->{relay} = join(',', @formatted_relays);

    my $action = join ',', map { ucfirst } split /,/, $self->{action};
    $self->{action} = $action;

    my @slices = @values[@COLS .. $#values];
    while (@slices) {
	my $pfqid = shift @slices;
	my $aseek = shift @slices || '-';
	my $zseek = shift @slices || '-';
	(my $afile = $aseek) =~ s/:\d+$//;
	(my $zfile = $zseek) =~ s/:\d+$//;

	if ($afile eq $zfile) {
	    push @{$self->{slices}}, [$pfqid, $aseek, $zseek];
	}
	else {
	    push @{$self->{slices}}, [$pfqid, $aseek, "$afile:999999999"];
	    push @{$self->{slices}}, [$pfqid, "$zfile:0", $zseek];
	}
    }
}

sub field {
    my ($self,$field) = @_;
    return $self->{$field};
}

sub logdata {
    my $self = shift;

    # Pre-filter actions, so that delivered messages don't show up as copies
    my @actions = grep { $_ ne 'Copy' } split(',', $self->{action});
    my @data = @{ $self->{mta}->logline_data(
                    reason => $self->{reason},
                    action => join(',', @actions),
                    rawlines => $self->_rawlog,
                    ) };

    $self->{$_} and $self->{$_} eq '-' and undef $self->{$_} for qw(to msgid);
   
    unshift @data, ['Message-ID' => "<$self->{msgid}>"] if $self->{msgid};
    unshift @data, ['Recipient(s)' => $self->{to}] if $self->{to};

    if ($self->{routes}) {
        foreach my $r (@{$self->{routes}}) {
            push @data, ['Routed' => "to $r"];
        }
    }
    if ($self->{copies}) {
        foreach my $r (@{$self->{copies}}) {
            push @data, ['Copied' => "to $r"];
        }
    }
    return \@data;
}

sub rawlog {
    my $self = shift;
    return $self->_rawlog(@_);
}

sub _rawlog {
    my $self = shift;
    die "Message was retrieved from a remote host and does not contain log data" if $self->{_no_log_data};
    my $gunzip_to_stdout = $PRODUCT->get_util('gunzip_to_stdout');
    my $gunzip = $PRODUCT->get_util('gunzip');
    my $gzip = $PRODUCT->get_util('gzip');
    my $tail_n = $PRODUCT->get_util('tail_n');
    
    return $self->{rawlog} if $self->{rawlog};

    return '' unless ref $self->{slices};

    my %lines;

    my $now = strftime('%Y-%m-%d %H:%M:%S', localtime);

    for my $slice (@{$self->{slices}}) {
	my $pfqid = $slice->[0];
	my ($afile,$aseek) = split ':', $slice->[1];
	my ($zfile,$zseek) = split ':', $slice->[2];

	die "$zfile != $afile" if $zfile ne $afile;
	
        my $log = "$Sophos::LogSearch::Config::config{path}->{indexes}/$afile";

	die "$log or $log.gz doesn't exist\n" unless -f $log or -f "$log.gz";

        # Actually retrieving the raw logs is not the common case, so we don't worry too 
        # much about the overhead of gunzipping and gzipping it afterward
        my $have_gunzipped = 0;
	unless (-f $log) {
          (system("$gunzip $log.gz") == 0) or die "Unable to extract $log.gz to $log\n";
          die "Couldn't create $log" unless -f $log;
          $have_gunzipped = 1;
	}

	my $cmd = qq{(cd $Sophos::LogSearch::Config::config{path}->{indexes} && echo "$pfqid:\t$aseek\t$zseek" | $self->{slicer} $log)};

	open my $fh, "$cmd |" or die "Can't execute $cmd: $!";
	while (my $line = <$fh>) {
	    my ($date,$rest) = $line =~ /(?:\d+:)?(\w+\s+\d+\s+\d{2}:\d{2}:\d{2}) (.*)/ or next;
	    $date = strftime('%Y-%m-%d %H:%M:%S', localtime(local_time_string_to_epoch($date)));
	    push @{$lines{$date}}, "$date $rest";
	}

	close $fh or die "$cmd returned bad exit status";
        if ($have_gunzipped && -f $log) {
            (system("$gzip $log") == 0) or die "Unable to gzip $log to $log.gz\n";
        }
    }

    $self->{rawlog} = join "\n", map { @{$lines{$_}} } sort keys %lines;
    return $self->{rawlog};
}

sub message_log_data {
    my ($self) = @_;
    die "Message was retrieved from a remote host and does not contain log data" if $self->{_no_log_data};

    return $self->{_message_log} if $self->{_message_log};

    $self->{milter_id} = $self->{mta}->get_milter_id($self->_rawlog);
    return unless $self->{milter_id};

    my $grep = $PRODUCT->get_util('egrep');
    my $message_log = $Sophos::LogSearch::Config::config{path}->{messagelog}; #$PRODUCT->get_file_location('messagelog');
    my $found = '';

    LOG:
    for my $log (sort <$message_log*>) {
        $found = qx{$grep $self->{milter_id} $log};
        last LOG if $found;
    }

    $self->{_message_log} = $found;
    return $self->{_message_log};
}

sub maillog_incoming {
    my ($self) = @_;
    
    return $self->{rawlog_incoming} if defined $self->{rawlog_incoming};

    $self->_divide_rawlog();

    return $self->{rawlog_incoming};
}

sub maillog_outgoing {
    my ($self) = @_;

    return $self->{rawlog_outgoing} if defined $self->{rawlog_outgoing};

    $self->_divide_rawlog();

    return $self->{rawlog_outgoing};
}

sub _divide_rawlog {
    my ($self) = @_;
    my $rawlog = $self->_rawlog;

    ($self->{rawlog_incoming}, $self->{rawlog_outgoing}) = $self->{mta}->get_divided_loglines($rawlog);

    return;
}

sub milter_host {
    my ($self) = @_;

    return $self->{milterhost} if defined $self->{milterhost};

    my $rawlog = $self->_rawlog;
    $self->{milterhost} = $self->{mta}->get_milter_host($rawlog);

    return $self->{milterhost}; 
}

sub split_data {
    my ($self) = @_;

    return $self->{split_data} if defined $self->{split_data};

    # Order of the to, aliases, actions, and reasons maps to each split, eg:
    #     @actions = ( 'split_1_action', 'split_2_action')
    #     @to      = ( 'split_1_to',     'split_2_to')
    # and so on.
    my @to = split(',', $self->{to});
                  # Take out 'Copy' actions, as they mess up the ordering
                  # mentioned above
    my @actions = grep { $_ ne 'Copy' } split(',', $self->{action});
    my @aliases = split(',', $self->{aliases});
    my @reasons = split(',', $self->{reason});

    my $splits;
    while (@to) {
        push @{$splits}, {
                    to     => shift @to || '-',
                    mapped => shift @aliases || '-',
                    action => shift @actions || '-',
                    reason => shift @reasons || '-',
                };
    }

    $self->{split_data} = $splits;
    return $splits;
}

1;
