package Sophos::LogSearch::Queue;

use strict;
use warnings;

use Sophos::LogSearch::Constants qw(&debug LOGSEARCH_DEBUG $P $MAMELOGD_QUEUE_SIZE);

my $LOGSEARCH_QUEUE_DEBUG = 0;

if (LOGSEARCH_DEBUG) {
    $LOGSEARCH_QUEUE_DEBUG = 1;
}

# ===============
# The queues

our $finished = [];          # Array ref for all the finished items
our $mail_id_queue = {};     # Holds information obtained from mail.log
our $milter_id_queue = {};   # Holds information obtained from message_log
our $queue_time = {};        # Queue bucketed by time slots

# ===============

sub new {
    my ($class,%args) = @_;

    my $self = {
        QUEUE_LIMIT           => 1000,  # max. number of objects to hold in the
                                        # queue at one time (in both the mail id
                                        # and milter queues)
        QUEUE_LIFE            => 120,   # in seconds

        %args,
        items_in_mail_queue   => 0,
        items_in_milter_queue => 0,
        items_in_queue_time   => 0,
        latest_timestamp      => 0,     # most recent timestamp of all added items
        latest_add_time       => time,  # time add() was most recently called
    };

    bless $self, $class;
    return $self;
}


# Add a transaction to the queue
# Transaction must have:
# - either a mail id or milter id (available through the id() or 
#   milter_id() methods)
# - a time (available through the time() method)
sub add {
    my ($self, $trans) = @_;
    $P->begin('queue add');
    my ($id, $milter_id, $id2, $time) = ($trans->id, $trans->milter_id, 
                                         $trans->id2, $trans->time);
    
    my $now = time; # XXX TODO: perf impact

    # We could be getting transactions out of order (in time)
    if (defined $time && ($time > $self->{latest_timestamp})) {

        $self->{latest_timestamp} = $time;
    }
    $self->{latest_add_time} = $now;

    # If this is a route-copy transaction, then it contains
    # all the information we need (1 line per mail)
    if ($trans->{contents}{route_copy}) {
        $self->mark_item_as_done($trans);
        $P->end('queue add');
        return;
    }

    # Mail ID:
    if ($id) {

        # If this mail id already exists, add the new information to the 
        # existing id
        if (exists $mail_id_queue->{$id}) {
            debug("Queue: add: id '$id' exists: merging");
            $self->merge_items($trans, $mail_id_queue->{$id});

        } else {
            # It's a new id, so add it to the queue
            $trans->add('queue_add_time' => $now);
            debug("Queue: add: creating entry for id '$id'");
            $mail_id_queue->{$id} = $trans;
            $self->{items_in_mail_queue}++;
            debug("Queue: adding transaction to bucket (time : '".($time||'')."' / id: '".($id||'')."')");
            $queue_time->{$time}->{$id} = $trans;
            $self->{items_in_queue_time}++;
        }
    }

    # Milter ID:
    if ($milter_id) {

        # Only add the data to the milter id queue if we haven't already
        # added it to the mail id queue

        if ($id) {
            # We've already added the data in $trans to the mail id queue, so
            # just merge over any data from the milter id queue
            if (exists $milter_id_queue->{$milter_id} and
                $milter_id_queue->{$milter_id} != $mail_id_queue->{$id} # They're not the same entry
            ) {
                # merge the existing item in milter_id_queue with $id in the mail_id_queue 
                # and link
                debug("Queue: add: id '$id' exists and milter id '$milter_id' exists: merging");
                $self->delete_item_from_queue_time($milter_id_queue->{$milter_id},$milter_id);
                $self->merge_items($milter_id_queue->{$milter_id}, $mail_id_queue->{$id});
                debug("Queue: add: linking milter id '$milter_id' to id '$id'");
                $self->link_items(\$milter_id_queue->{$milter_id}, \$mail_id_queue->{$id});

            } else {
                # Just create the link in the milter_id queue (not merge)
                $milter_id_queue->{$milter_id} = {};   # Create a dummy entry
                debug("Queue: add: linking milter id '$milter_id' to id '$id'");
                $self->link_items(\$milter_id_queue->{$milter_id}, \$mail_id_queue->{$id});
            }

        } else { # No mail ID

            # If milter id already exists, merge the two
            if (exists $milter_id_queue->{$milter_id}) {
                # merge $trans with the existing item in the queue
                debug("Queue: add: milter id '$milter_id' exists and no mail id: merging");
                $self->merge_items($trans, $milter_id_queue->{$milter_id}); 

            } else {
                # It's a new ID, so create it
                $trans->add('queue_add_time' => $now);
                debug("Queue: add: creating entry for milter id '$milter_id'");
                $milter_id_queue->{$milter_id} = $trans;
                $self->{items_in_milter_queue}++;
                die "undefined: time $time" unless $time;
                die "undefined: id $milter_id" unless $milter_id;
                die "Duplicate entry for $milter_id" if exists($queue_time->{$time}->{$milter_id});
                $queue_time->{$time}->{$milter_id} = $trans;
                $self->{items_in_queue_time}++;
            }

        }
    }

    # Deal with the second mail ID - we should only get one if we also have
    # the first ID
    if ($id2 and $id) {

        if (exists $mail_id_queue->{$id2} and
            $mail_id_queue->{$id} != $mail_id_queue->{$id2} # They're not the same entry
            ) {
            # Merge it into id1 and link id2 -> id1

            debug("Queue: add: id '$id' exists and id2 '$id2' exists: merging");
            $self->delete_item_from_queue_time($mail_id_queue->{$id2},$id2);
            $self->merge_items($mail_id_queue->{$id2}, $mail_id_queue->{$id});
            debug("Queue: add: linking '$id2' to '$id'");
            $self->link_items(\$mail_id_queue->{$id2}, \$mail_id_queue->{$id});

        } else {
            # Just link the entry for ID2 to the entry for ID1

            debug("Queue: add: linking '$id2' to '$id'");
            $mail_id_queue->{$id2} = 1;    # Create a dummy entry
            $self->link_items(\$mail_id_queue->{$id2}, \$mail_id_queue->{$id});

        }

    }



    # Now decide whether or not the transaction is done
    if (defined($id) && $mail_id_queue->{$id}->done) {
        $self->mark_item_as_done($mail_id_queue->{$id});

    } elsif (defined($milter_id) && $milter_id_queue->{$milter_id}->done) {

        $self->mark_item_as_done($milter_id_queue->{$milter_id});
    }
    $P->end('queue add');

    return;
}


sub mark_item_as_done {
    my ($self, $trans) = @_;

    return if $trans->contents->{finished};

    if ($LOGSEARCH_QUEUE_DEBUG) {
        my $id = $trans->id || 'undef';
        my $id2 = $trans->id2 || 'undef';
        my $milter_id = $trans->milter_id || 'undef';
        debug("Queue: done item (id: '$id', id2: '$id2', milter id: '$milter_id')");
    }
    $trans->add('finished' => 1);
    push @{$finished}, $trans;
}

sub finalize {
    my ($self, $trans) = @_;
    my $c = $trans->contents;

    my %scache;
    my %status = %{ $c->{STATUS} || {} };
    # outgoing overrides incoming
    for my $qid ($trans->id, $trans->id2) {
        if ($qid && exists $status{$qid}) {
            my $s = delete $status{$qid};
            $scache{$_} = $s->{$_} for keys %$s;
        }
    }

    my %failed = map {
        $scache{$_} ne 'sent' ? ($_ => $scache{$_}) : ()
    } keys %scache, keys %status;

    my $action;
    my $found = 0;
    if ($trans->done_message_log && exists $c->{ACTION} && exists $c->{TO}) {
        my @acts = split ',', $c->{ACTION};
        my @addr = split ',', $c->{TO};

        if (scalar @addr eq scalar @acts) {
            my @nacts;
            for my $i (0 .. $#addr) {
                my $act = $acts[$i];
                if (exists $failed{$addr[$i]}) {
                    $act = $failed{$addr[$i]} eq 'deferred'
                           ? 'deferred' : 'bounced';
                    $found ++;
                }
                else {
                    $act = 'bounced' if $act eq 'reject';
                }
                push @nacts, $act;
            }
            $action = join ',', @nacts;
        }
    }

    # Unaccounted for failures - Mark as globally failed
    if ($found != keys %failed) {
        $action = 'deferred';
        if (grep { /(?:bounced|undeliverable)/ } values %failed) {
            $action = 'bounced';
        }
    }

    $trans->add(ACTION => $action) if $action;
    return;
}

sub delete_item_from_queue_time {
    my ($self,$trans,$id) = @_;
    return if !defined $id;
    my $time = $trans->time;
    return unless exists $queue_time->{$time}->{$id};
    debug("Queue: Deleting transaction with id '$id' from the time queue (in bucket '$time')");
    delete($queue_time->{$time}->{$id});
    $self->{items_in_queue_time}--;
}

# Deletes an item from all queues
sub delete_item {
    my ($self, $trans) = @_;

    my ($id, $milter_id, $id2) = ($trans->id, $trans->milter_id, $trans->id2);

    debug("Queue: Deleting item (mail id: '".($id || '')."', milter id: '". ($milter_id||'')."', id2: '".($id2||'') . "'" );

    # Remove the original item and all links to it (eg: second mail id or 
    # milter id)
    if (defined $id && exists $mail_id_queue->{$id}) {
        delete $mail_id_queue->{$id};
        $self->{items_in_mail_queue}--;
    }

    if (defined $id2 && exists $mail_id_queue->{$id2}) {
        delete $mail_id_queue->{$id2};
    }

    if (defined $milter_id && exists $milter_id_queue->{$milter_id}) {
        delete $milter_id_queue->{$milter_id};
    }

    # Only in milter queue
    if (defined($milter_id) && !defined($id)) {
        $self->{items_in_milter_queue}--;
    }

    return;
}


# Merge two transaction objects - the resulting merged object is
# stored in transaction #2
sub merge_items {
    my ($self, $from_trans, $to_trans) = @_;

    my $from = $from_trans->contents;
    my $to = $to_trans->contents;

    # use Data::Dumper;
    # debug("merge_items: from before: " . Dumper($from) );
    # debug("merge_items: to before: " . Dumper($to) );

    if ($LOGSEARCH_QUEUE_DEBUG) {
        my $from_id1 = $from_trans->id || 'undef';
        my $to_id1 = $to_trans->id || 'undef';
        my $from_milter_id = $from_trans->milter_id || 'undef';
        my $to_milter_id = $to_trans->milter_id || 'undef';
        my $from_id2 = $from_trans->id2 || 'undef';
        my $to_id2 = $to_trans->id2 || 'undef';

        debug("Queue: Merging (id1: '$from_id1', id2: '$from_id2', milter id: '$from_milter_id') with " .
              "(id1: '$to_id1', id2: '$to_id2', milter id: '$to_milter_id')");
    }

    # We always prefer message_log data over mail.log data for some keys,
    # because to, alias, group, action and reason are *ordered* lists,
    # and the message_log data gives them in the right order.  We risk losing
    # that order if we take the mail.log data

    my @message_log_override_keys = (qw/REASON GROUPS ALIASES TO/);

    if ($from_trans->done_message_log && !$to_trans->done_message_log) {

        foreach my $key (@message_log_override_keys) {
            if ($from->{$key}) {
                $to->{$key} = $from->{$key};
                delete $from->{$key};
            }
        }

    } elsif (!$from_trans->done_message_log && $to_trans->done_message_log) {

        foreach my $key (@message_log_override_keys) {
            if ($from->{$key}) {
                # Just delete the from key - we'll keep the to transaction's 
                # message_log info
                delete $from->{$key};
            }
        }
    }

    for my $key (keys %$from) {

        next if !defined($from->{$key});

        if ( ($key eq 'mail_log') || ($key eq 'message_log') ) {
            # Just concatenate maillogs
            $to->{$key} .= $from->{$key};

        } elsif ( ( ($key eq 'id') || ($key eq 'milter_id') ) && (exists $to->{$key}) ) {
            # Always take the destination id
            next;
        
        } elsif ($key eq 'done_mail_log') {
            $to->{$key}++;

        } elsif (is_blank($to->{$key})) {
            $to->{$key} = $from->{$key} unless is_blank($from->{$key});

        } elsif ($key eq 'STATUS') {
            if (exists $from->{id}) {
                my $id = $from->{id};
                $to->{$key} ||= {};
                for my $qid (keys %{$from->{$key}}) {
                    for my $status (keys %{$from->{$key}{$qid}}) {
                        $to->{$key}{$qid}{$status} = $from->{$key}{$qid}{$status};
                    }
                }
            }
        } elsif (ref($from->{$key}) eq 'HASH') {
            $to->{$key} = merge_hash($to, $from, $key);

        } elsif (ref($to->{$key}) eq 'HASH') {
            $to->{$key} = merge_hash($from, $to, $key);

        } elsif ( ($to->{$key} eq $from->{$key}) or is_blank($from->{$key}) ) {
            next;

        } elsif ($key eq 'ETIME_EPOCH') {
            $to->{ETIME_EPOCH} = $from->{ETIME_EPOCH} if $from->{ETIME_EPOCH} lt $to->{ETIME_EPOCH};

        } elsif ($key eq 'ATIME_EPOCH') {
            $to->{ATIME_EPOCH} = $from->{ATIME_EPOCH} if $from->{ATIME_EPOCH} lt $to->{ATIME_EPOCH};

        } elsif ($key eq 'ATIME') {
            $to->{ATIME} = $from->{ATIME} if $from->{ATIME} lt $to->{ATIME};

        } elsif ($key eq 'queue_add_time') {
            # Always take the newer time
            my ($highest) = reverse sort ($to->{queue_add_time}, $from->{queue_add_time});
            $to->{$key} = $highest;

        } elsif ($key =~ /^(?:CLIENT|RELAY)$/) {
            $to->{$key} = $from->{$key} if $to->{$key} =~ /^(local|127\.0\.0\.1)$/;

        } else {
            $to->{$key} .= ",$from->{$key}";
        }
    }

    $to_trans->set_contents($to);
}

sub merge_hash {
    my ($from, $to, $key) = @_;

    if (!exists $from->{$key}) {
        $from->{$key} = {};

    } elsif (ref($from->{$key}) ne 'HASH') {
        $from->{$key} = { $from->{$key} => 1 };
    }

    return { (%{$to->{$key}}, %{$from->{$key}}) };
}

sub link_items {
    my ($self, $from_ref, $to_ref) = @_;

    $$from_ref = $$to_ref;
}


sub is_blank {
    my $val = shift;
    return (
       not $val
        or $val eq '-'
        or (ref $val eq 'HASH' and not %$val)
        or (ref $val eq 'ARRAY' and not @$val)
    ) ? 1 : 0;
}


# Returns the list of all finished items and removes them from all queues
sub finished_items {
    my ($self) = @_;

    if (scalar(@{$finished}) ) {
	foreach my $trans (@{$finished}) {
            if ($LOGSEARCH_QUEUE_DEBUG) { 
                debug("Queue: finishing item (mail id: '".($trans->id || '')."', milter id: '". ($trans->milter_id||'')."', id2: '".($trans->id2||'') . "'" );
            }
            $self->finalize($trans);
            $self->delete_item($trans);
            $self->delete_item_from_queue_time($trans, $trans->id);
        }

        my $finished_items = $finished;
        $finished = [];
        return $finished_items;

    } else {

        # No items, so return undef
        return undef;

    }
}


# Return the oldest finished item from the finished queue
sub next_finished_item {
    my ($self) = @_;

    if (scalar(@{$finished}) ) {

        my $trans = shift @{$finished};
        $self->delete_item($trans);
        $self->delete_item_from_queue_time($trans, $trans->id);
        return $trans;

    } else {

        # No items, so return undef
        return undef;

    }
}


sub next_old_enough_finished_item {
    my $self = shift;

    if (scalar(@{$finished})) {

        my $oldest_time = $finished->[0]->contents->{queue_add_time};
        if ( (time - $oldest_time) > $self->{QUEUE_LIFE} ) {

            return $self->next_finished_item;
        }

    }

    return undef;

}


sub unfinished_items {
    my ($self) = @_;

    my @vals;
    my %seen_ids;

    foreach my $trans (values(%{$mail_id_queue}), values(%{$milter_id_queue}) ) {
        my $id = $trans->id;
        my $milter_id = $trans->id;
        # Add this transaction to the unfinished list if:
        # 1) We haven't seen its ID before
        # 2) It only has a milter id (and is therefore unique)
        # 3) AND it thinks it's not done
        if ( !$trans->done &&
             (!$id || !exists $seen_ids{$id}) &&
             (!$milter_id || !exists $seen_ids{$milter_id})
           ) {
            push @vals, $trans;
            $seen_ids{$id} = 1 if $id;
            $seen_ids{$milter_id} = 1 if $milter_id;
        }
    }

    if (scalar(@vals)) {
        return \@vals;

    } else {
        return undef;
    }
}


# Return the transaction with the given mail id or milter id
sub search {
    my ($self, $id) = @_;

    if (exists $mail_id_queue->{$id}) {

        return $mail_id_queue->{$id};

    } elsif (exists $milter_id_queue->{$id}) {

        return $milter_id_queue->{$id};

    } else {
        return undef;
    }

}

# Delete all items in all queues
sub delete_all {
    my $self = shift;

    $finished = [];
    $queue_time = {};
    $self->{items_in_queue_time} = 0;
    $mail_id_queue = {};
    $milter_id_queue = {};
    $self->{items_in_milter_queue} = 0;
    $self->{items_in_mail_queue} = 0;
    $self->{latest_timestamp} = 0;
}


# ===============
# Queue status functions


# Returns the number of items in the queue
sub item_count {
    my $self = shift;

    return $self->{items_in_queue_time};
}


# Returns whether or not we're over the queue limit
sub too_full {
    my $self = shift;

    return $self->item_count > $self->{QUEUE_LIMIT};
}


# ===============
# Time-related functions


# Return the newest timestamp in the queue
sub time_of_newest_item {
    my $self = shift;

    return $self->{latest_timestamp} || 0;
}


# Returns the time of the most recent call to add()
sub latest_add_time {
    my $self = shift;
    
    return $self->{latest_add_time};
}

sub _flush {
    my ($self,$limit) = @_; 

    my $buckets = [];
    my @remove_these = 
       sort { $a <=> $b } grep { $_ < $limit } keys %$queue_time;
    return unless @remove_these;
    my $deleted_transactions = [];
    push(@$buckets,@{$queue_time}{@remove_these});
    foreach my $bucket (@$buckets) { 
       foreach my $val (values %$bucket) {
           $self->{items_in_queue_time}--; #if $val;
           $self->delete_item($val) if $val;
           push @$deleted_transactions, $val;
       }
   }
   delete(@{$queue_time}{@remove_these});
   return $deleted_transactions;
}

 

# Remove any items from the queue that we no longer care about:
# - Items older than the latest timestamp
# - If there are more than QUEUE_LIMIT items in the queue, remove
#   the oldest items until we have (QUEUE_LIMIT - 1) items left
sub flush {
    my $self = shift;
    my $iter =0;
    my $limit = $self->{latest_timestamp} - $self->{QUEUE_LIFE};
    if ($LOGSEARCH_QUEUE_DEBUG) {
        debug("Queue: flush (items in queue: " . $self->{items_in_queue_time} . " / latest timestamp: " . $self->{latest_timestamp} . ")");
    }

    my $deleted_transactions = [];
    while($self->too_full){
       my $deleted = $self->_flush($limit);
       push @$deleted_transactions, @$deleted if $deleted;
       $limit++;
       $iter++;
       if ($iter > 501) {
          # something has gone wrong!
          use Data::Dumper;
          die "QUEUE" . Dumper($queue_time) . "\n count: $self->{items_in_queue_time} limit: $limit";
          $self->{items_in_queue_time} = 0 unless keys %$queue_time;
       }
    }
    return $deleted_transactions;
}


# Returns whether or not a transaction is in the mail id queue and 
# milter queue
sub not_in_queues {
    my ($self, $trans) = @_;

    my $id = $trans->id;
    my $milter_id = $trans->milter_id;

    # Exists in the queue if:
    #    we have an id and it's not in the mail id queue
    #       - and -
    #    we have a milter id and it's not in the milter id queue

    return ( (!$id || !exists $mail_id_queue->{$id}) && 
        (!$milter_id || !exists $milter_id_queue->{$milter_id}) );
}

1;
