package Sophos::LogSearch::Source::Indexer;

use strict;
use warnings;

use Carp;
use ActiveState::Handy           qw(run);
use List::Util                   qw(first);
use File::Basename               qw(dirname basename);
use File::Path                   qw(mkpath rmtree);
use POSIX                        qw(strftime);
use IO::Handle;
use Hash::Dynamic;
use Storable qw(store retrieve);
use ActiveState::DiskUsage       qw(du);

use Sophos::LogSearch::Transaction;
use Sophos::LogSearch::Config;
use Sophos::LogSearch::Constants qw($PREFIX $PRODUCT $DEFER_DAYS 
                                    %ALL &debug $MAMELOGD_QUEUE_SIZE &offset_format
                                    &mame_log utc_time_string_to_epoch format_date);


# ===============
# Constants

my %OPEN_INDEXES;   # Holds all open index filehandles
my @time_range = ();
my $warning_oversize = 0; # warning about oversized log directories.

my $mamelog;
    # {lines}       Lines written to current seg (rolls at 1000).
    # {path}        "$self->{BASEDIR}/yyyy-mm-dd/mamelog"
    # {seg_start}   Start offset in mamelog for current seg.
    # {tell}        Current output offset
    # {time}        Last time written; "yyyy-mm-dd hh:mm:ss"
    # {ts}          {time} as UTC epoch.
    # {maillog}     "maillog file"

my $hostname    = $PRODUCT->hostname();
my $fqdomain    = $PRODUCT->fqdn();


sub gzip    { croak "file required" unless $_[0]; my $gzip = $PRODUCT->get_util('gzip'); run("\@$gzip $_") for @_ }

sub new {
    my ($class, %args) = @_;

    my $self = {

        QUEUE_LIMIT   => $MAMELOGD_QUEUE_SIZE,
        DEFER_DAYS    => $DEFER_DAYS,
        BASEDIR       => $Sophos::LogSearch::Config::config{path}->{indexes},
        MONITOR_LOCAL_TRAFFIC => $Sophos::LogSearch::Config::config{index_local_delivery},
        PER_DAY       => 10,      # Lengths of prefixes of yyyy-mm-ddThh:mm:ss
        LINES_PER_INDEX => 5000,  # Number of lines (records) per index file
        LINES_PER_MAILLOG_INDEX => 50_000,  # Number of lines (records) per maillog

        # Internal variables:
        dir           => undef,   # Current directory we're writing indexes to 
                                  #    ("$self->{BASEDIR}/yyyy-mm-dd")
        'tell'        => 0,       # Current offset into the main index (its tell)
        maillog_tell  => 0,       # Current offset into the maillog (its tell)
        main_index    => undef,   # The current opened main index file 
                                  # (e.g. 2006-11-23/mamelog.123456789)
        main_index_fh => undef,   # Filehandle for the current main index file
        lines_written => 0,       # Number of lines that have been written to the
                                  #   currently open main index file
        maillog_lines_written => 0,       # Number of lines that have been written to the
                                  #   currently open main index file
        lines_total   => 0,       # total lines written
        %args,

    };

    $self->{resume_file}   = "$self->{BASEDIR}/resume.dat";
    $self->{stop_time_file}   = "$self->{BASEDIR}/resume.stoptime";
    $self->{log_resume_file} = "$self->{BASEDIR}/resume.loglines";

    return bless $self, $class;
}


sub init {
    my $self = shift;

    #-- Get mame_time: last timestamp of last non-empty mamelog.
    my $latest_time = $self->get_newest_time;
    $mamelog->{time} = $latest_time;

    $self->{dir} ||= '';
    return unless $mamelog->{time};

    # Prune @maillogs/@maillogs by searching backward for $mamelog->{time}
    $mamelog->{ts} = utc_time_string_to_epoch($mamelog->{time});
    debug("Indexer: mamelog init: ts='$mamelog->{ts}', time='$mamelog->{time}'");

    return $mamelog->{time};
}


# Gets the latest time written to the index files
sub get_newest_time {
    my $self = shift;

    my $tail_n = $PRODUCT->get_util('tail_n');
    my $gunzip_to_stdout = $PRODUCT->get_util('gunzip_to_stdout');
    my $sort = $PRODUCT->get_util('sort');
    my $awk = $PRODUCT->get_util('awk');

    my $latest_dir = ('',grep {-s} sort glob "$self->{BASEDIR}/20??-??-??")[-1];

    my @times;
                                                    # NumZeros
    foreach my $logfile ( ( glob("$latest_dir/mamelog.?????????[0-9]{,.gz}"), $self->{stop_time_file} )) {

        my $valid_awk_re =  q|'{ if($1 ~ /^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]:[0-9][0-9]:[0-9]/) print $1 }'|;
        next if !-e $logfile;
        my $cmd = $logfile =~ /\.gz$/
                    ? "$gunzip_to_stdout $logfile"
                    : "cat $logfile";
        $cmd .= " | $awk $valid_awk_re | $sort | $tail_n 1";

        my $latest_dir_time = qx($cmd);
        push @times, $latest_dir_time if $latest_dir_time;
    }

    return unless @times;
    my $latest_time = (sort @times)[-1];

    chomp $latest_time;
    return $latest_time;
}


# Format and output the timestamp to our resume file
sub save_stop_time {
    my ($self, $stop_time) = @_;

    my $date = format_date(gmtime($stop_time));
    open my $FH, ">$self->{stop_time_file}" or die "Can't open '$self->{stop_time_file}' for open: $!";
    print $FH "$date\n";
    close $FH or die "Error closing file '$self->{stop_time_file}': $!";
}


# Load the 'in-flight' transaction data from disk
sub restore_transaction_data {
    my $self = shift;

    return unless -e $self->{resume_file};
    my $data;

    eval {
      $data = retrieve($self->{resume_file});
      unlink($self->{resume_file});
    };
    return $data unless $@;
    return;
}


# Save the 'in-flight' transaction data to disk
sub save_transaction_data {
    my ($self,$data) = @_;
    store($data, $self->{resume_file});
}


# Save the last maillog and message_log lines to disk
sub save_log_lines {
    my ($self, $maillog_line, $messagelog_line,$maillog_seek,$messagelog_seek) = @_;

    # If we haven't read anything more since resuming
    # don't overwrite what we currently think where we should be
    # resuming to. 
    return if !(defined $maillog_line && defined $maillog_seek);

    debug("Indexer: save_log_lines: maillog line: '" . ($maillog_line || '')
          . "' / messagelog line: '" . ($messagelog_line || '')
          . "' / maillog seek: '" . ($maillog_seek || '') ."'");
    open my $FH, ">$self->{log_resume_file}" or die "Error opening '$self->{log_resume_file}': $!";
    print $FH "maillog: $maillog_line\n" if defined $maillog_line;
    print $FH "messagelog: $messagelog_line\n" if defined $messagelog_line;
    print $FH "maillog_seek: $maillog_seek\n" if defined $maillog_seek;
    print $FH "messagelog_seek: $messagelog_seek\n" if defined $messagelog_seek;
    close $FH or die "Error closing '$self->{log_resume_file}': $!";
}


# Restore the saved maillog and message_log lines from disk
sub restore_log_lines {
    my $self = shift;

    return if !-e $self->{log_resume_file};

    open my $FH, "<$self->{log_resume_file}" or die "Error opening '$self->{log_resume_file}': $!";
    my $contents = do { local $/; <$FH> };
    close $FH or die "Error closing '$self->{log_resume_file}': $!";

    my $lines = {};
    ($lines->{mail_log})    = $contents =~ m{maillog: (.*)$}m;
    ($lines->{message_log}) = $contents =~ m{messagelog: (.*)$}m;
    ($lines->{maillog_seek}) = $contents =~ m{maillog_seek: (.*)$}m;
    ($lines->{messagelog_seek}) = $contents =~ m{messagelog_seek: (.*)$}m;
    return $lines;
}


sub calculate_tell {
    my ($self, $full_path) = @_;
    my $gunzip_info = $PRODUCT->get_util('gunzip_info');
    my $awk = $PRODUCT->get_util('awk');

    my $tell = 0;
					     # NumZeros
    if (my $last_seg = (sort glob("$full_path.[0-9]????????[0-9]{,.gz}"))[-1]) {

        debug("Indexer: searched for and found last segment file of '$last_seg'");

                                # NumZeros
        ($tell) = $last_seg =~ /\.(\d{10})/;

        if ($last_seg =~ /\.gz$/) {

            my $size = qx/$gunzip_info $last_seg | $awk 'NR==2{print \$2}'/;
            die "Error finding file size" unless defined $size;
            debug("Indexer: unzip doing a $gunzip_info $last_seg");
            $tell += $size;

        } else {
            my $size = (stat $last_seg)[7];
            $tell += $size;
        }

    }

    debug("Indexer: calculated new tell for '$full_path': $tell"); 
    return $tell;
}


sub calculate_mamelog_tell {
    my $self = shift;
    $mamelog->{path} = shift;

    $self->{tell} = $self->calculate_tell($mamelog->{path});

    return; 
}


sub close_current_segment {
    my ($self, %args) = @_;

    # We sometimes don't want to gzip the segment (if we're only shutting
    # down, not rolling over to another day, for example)
    my $gzip_it = $args{no_gzip} ? 0 : 1;

    debug("Indexer: close_mamelog_seg has a count of " . ($self->{lines_written} || 0) );
    if ($self->{lines_written}) {
        mame_log("close_mamelog_seg $self->{main_index} at $self->{lines_written}");
        debug("Indexer: closing segment $self->{main_index} with $self->{lines_written} written to it.");
        close $self->{main_index_fh};
        $self->log_mamelog_time(@time_range);

        if ($gzip_it) {
            mame_log("gzip $self->{main_index}");
            gzip $self->{main_index};
        }
    }

    $self->{lines_written} = 0;
    @time_range = ();
    return;
}


sub put_mamelog_seg {
    my ($self, $line) = @_;

    # size of mamelog.000 files
    if ($self->{lines_written} == $self->{LINES_PER_INDEX}) {
        $self->close_current_segment;
        $self->{lines_written} = 0;
    }

    if ($self->{lines_written} == 0) {
        $mamelog->{seg_start} = offset_format($self->{tell});
        $self->{main_index} = "$self->{dir}/mamelog.$mamelog->{seg_start}";
        $self->{main_index_fh} = $self->open_out($self->{main_index});
        $self->{tell} = $self->calculate_tell("$self->{dir}/mamelog");
    }

    my $time = substr($line,0,19);
    $time_range[0] ||= $time;
    $time_range[1] = $time;

    $self->{main_index_fh}->print($line);
    $self->{tell} += length($line);
    ++$self->{lines_written};
    return;
}


sub put_maillog {
    my ($self, $content) = @_;

    my $gunzip = $PRODUCT->get_util('gunzip');

    # size of maillog.000 files
    if ($self->{maillog_lines_written} > $self->{LINES_PER_MAILLOG_INDEX}) {
        $self->close_current_maillog;
        $self->{maillog_lines_written} = 0;
        $self->{maillog_tell} = 0;
    }

    if ($self->{maillog_lines_written} == 0) {
        $self->{maillog_tell} = 0;
        my $offset = $self->calculate_tell("$self->{dir}/maillog");
        my $formatted_tell = offset_format($offset);
        $mamelog->{maillog} = "$self->{dir}/maillog.$formatted_tell";
        # Just in case
        if (-f "$mamelog->{maillog}.gz") {
            run("\@$gunzip $mamelog->{maillog}.gz");
        }
        $mamelog->{maillog_fh} = $self->open_out($mamelog->{maillog});
    }

    my $end_line_seek;
    my $start_line_seek = $end_line_seek = $self->{maillog_tell};
    foreach my $line (split "\n", $content) {
        my $raw = "$line\n";
        $end_line_seek = $self->{maillog_tell};
	$mamelog->{maillog_fh}->print($raw);
	$self->{maillog_tell} += length($raw);
	++$self->{maillog_lines_written};
    }
    # offset of beginning of first line
    # offset of beginning of last line
    return ($start_line_seek, $end_line_seek);
}

#-------------------------------------------------------------------------------
# Open an output file(append, autoflush)

sub open_out {
    my  ($self,$file) = @_;
    #XXX ensure dirname($file) exists.
    use Carp qw/croak/;
    open(my $fh, '>>', $file) or croak "Cannot open $file for output: $!";
    $fh->autoflush(1);
    #chown $pmx, 0, $file or die "Cannot chown pmx $file: $!";
    debug("Indexer: opening index file '$file'");
    return $fh;
}

sub log_mamelog_time {
    my ($self,@time_range) = @_;
    my $file = "$mamelog->{path}.time";
    mame_log("log_mamelog_time $mamelog->{seg_start} @time_range >> $file");
    debug("Indexer: log_mamelog_time: opening/creating new time index '$file'");
    if (@time_range) {
        my $time_idx = $self->open_out($file);
        print $time_idx "$mamelog->{seg_start} @time_range\n";
        debug("Indexer: emit to index ('$file'): $mamelog->{seg_start} @time_range");
        close $time_idx;
    }
    return;
}


# Write next transaction to main log and indexes
sub write {
    my ($self, $trans) = @_;
 
    debug("Indexer: writing transaction with mail id " . ($trans->id ? $trans->id : "undef") );

    my $mail_contents = $trans->contents;

    my $time = $mail_contents->{ATIME};
    unless($time) {
        # this transaction has gone through messagelog only, so it doesn't have an ATIME.  We still need
        # something to index on, so use ETIME instead
        $time = $mail_contents->{ATIME} = $mail_contents->{ETIME};
    }
    mame_log("mame: $mamelog->{time} > $time") if ( defined($mamelog->{time}) && ($mamelog->{time} gt $time) );
    $mamelog->{time} = $time;

    $self->create_new_indexes($time, $trans);

    # Write out the raw logs
    my $slices = '';
    if (exists $mail_contents->{mail_log}) {
        my ($s_offset, $e_offset) = $self->put_maillog($mail_contents->{mail_log});

        # Slicer expects a non-absolute path
        (my $maillog = $mamelog->{maillog}) =~ s#$self->{BASEDIR}/##;
    
        # Construct slices used to find the raw log lines.  They should look like:
        #     MAILID path/to/file:tell1 path/to/file:tell2 

        if (my $id1 = $trans->id) {
            $id1 = 'NOQUEUE' if $id1 =~ /^NOQUEUE/;
            $slices = "$id1 $maillog:$s_offset " .
                      "$maillog:$e_offset";
        }
        
        if (my $id2 = $trans->id2) {
            $id2 = 'NOQUEUE' if $id2 =~ /^NOQUEUE/;
            $slices .= " $id2 $maillog:$s_offset " .
                       "$maillog:$e_offset";
        }
    }

    # We use the mamelog tell as the unique id for the record, which is 
    # written to both the mamelog file and the indexes
    my $tell = offset_format($self->{tell});

    # Write out the field and index data
    my $to = $trans->field('TO');
    if ($self->_should_fingerprint_email($trans)) {

        debug("Indexer: Writing log data for a message with to:'$to'");
        my $mamelog_line = '';
        foreach my $field ( @{$trans->get_fields} ) {
            $mamelog_line .= $field . ' ';
        }

        $self->put_mamelog_seg($mamelog_line . "$tell $slices\n"); #SIC!

        my $indexes = $trans->get_indexes;
        for my $key (keys %{$indexes}) {
            my $value = $indexes->{$key}; 
            $OPEN_INDEXES{$key}->print("$tell $value\n");
        }

	$self->{lines_total}++;

    } else {
        debug("Indexer: Not writing log data for a message with to:'$to'");
    }
    
    return;
}

sub statistics {
  my $self = shift;
  return ( lines => $self->{lines_total} );
}

sub create_new_indexes {
    my ($self, $time, $trans) = @_;

    my $gunzip  = $PRODUCT->get_util('gunzip');
    (my $day = substr($time,0,$self->{PER_DAY})) =~ tr{T}{/};
    my $dir = "$self->{BASEDIR}/$day";

    debug("Indexer: Calculated a dir we'd expected our data to be in: '$dir'");
   
    if (!defined($self->{dir}) || ($self->{dir} ne $dir)) {	# End-of-day
        
        my $indexes = $trans->index_names;
        debug("Indexer: Need to rotate our data dir from '$self->{dir}'") if defined $self->{dir};

        if ($self->{dir}) {
            $self->close_all_indexes;
        }

        $self->{dir} = $dir;
        if (!-d $dir) {
            mkpath($dir);
        }
        
        # REOPEN gzipped indexes !? implies clock shifted back one day.
        if (my $gz = join " ", grep {-e && -f} map {"$self->{dir}/mamelog.$_.gz"} @{$indexes}) {
            mame_log("Unzipping old indexes: '$gunzip $gz'");
            run("\@$gunzip $gz");
            debug("Indexer: Unzipping old indexes: '$gunzip $gz'");
        }
        $self->calculate_mamelog_tell("$self->{dir}/mamelog");

        $OPEN_INDEXES{$_} = $self->open_out("$self->{dir}/mamelog.$_") for @{$indexes};

        my $tell = $self->calculate_tell("$self->{dir}/maillog");
        my $formatted_tell = offset_format($tell);
        $mamelog->{maillog} = "$self->{dir}/maillog.$formatted_tell";

        if (-f "$mamelog->{maillog}.gz") {
            run("\@$gunzip $mamelog->{maillog}.gz");
        }

        $mamelog->{maillog_fh} = $self->open_out($mamelog->{maillog});
        $self->{maillog_tell} = $self->{maillog_lines_written} = 0;

        mame_log("Start new log directory: '$self->{dir}'");
    }
}


# Close and gzip all indexes and our copy of the maillog
sub close_all_indexes {
    my $self = shift;

    my $gzip = $PRODUCT->get_util('gzip');

    # Close the main logfile
    $self->close_current_segment;

    close $self->{main_index_fh} if defined $self->{main_index_fh};

    # Close the indexes
    for my $key (grep { defined $OPEN_INDEXES{$_} } keys %OPEN_INDEXES) {
        $OPEN_INDEXES{$key}->close;
        system("$gzip $self->{dir}/mamelog.$key");
        debug("Indexer: gzipping index ($key): $self->{dir}/mamelog.$key");
    }

    %OPEN_INDEXES = ();

    # Close the raw mail logs file
    $self->close_current_maillog;

    return;
}


# Expire old index directories.  $days is the number of days to *keep*.
# Everything else is removed.
sub expire_old_indexes {
    my ($self, $days) = @_;

    my @indexes = grep {-d} sort glob "$self->{BASEDIR}/2???-??-??";

    my $to_expire = scalar(@indexes) - $days - 1;

    foreach my $dir (@indexes[0..$to_expire]) {
        # Rename the directory so that it can't be searched while we're doing
        # the rmtree
        _rename_and_delete_dir($dir);
    }
    return 1;
}

sub close_current_maillog {
    my $self = shift;

    my $gzip = $PRODUCT->get_util('gzip');

    if ($mamelog->{maillog_fh}) {
        close $mamelog->{maillog_fh};
        $mamelog->{maillog_fh} = undef;
    }
    if ($mamelog->{maillog}) {
        gzip($mamelog->{maillog});
        $mamelog->{maillog} = undef;
    }
    return;
}

# Returns size of all indexes on disk
sub total_index_size {
    my $self = shift;

    $self->{index_disk_space} = du($self->{BASEDIR});
    return $self->{index_disk_space};
}


# Prunes indexes on disk until their total size is less than $max_disk
sub prune_disk_space {
    my ($self, $max_disk) = @_;
    my $current_usage = $self->{index_disk_space};

    debug("Indexer: prune_disk_space");
    while($current_usage > $max_disk) {

       my @dirs = grep {-d} sort glob "$self->{BASEDIR}/2???-??-??";
       if (scalar @dirs) {
           # Dont prune today & the current working day (if different)
           @dirs = splice @dirs, 0, -1; 
           @dirs = grep {$_ ne $self->{dir}} @dirs if $self->{dir};
       }

       if (scalar @dirs) { 
           debug("Indexer: dirs to prune: " . join(', ', @dirs) . ")"); 

           my $harvest_dir = shift(@dirs);
           mame_log("max_disk_usage exceeded, attempting to harvest old index '$harvest_dir'.");
           debug("Indexer: pruning directory $harvest_dir");

           _rename_and_delete_dir($harvest_dir) or die <<EOF;

Logsearch max_disk_usage exceeded, but there was an error unlinking the directory '$harvest_dir', exiting.

EOF

           mame_log("Unlinked old index '$harvest_dir'.");
           $current_usage = $self->total_index_size();
           $warning_oversize = 0;
       } else {
           unless ($warning_oversize) {
              debug("Indexer:WARNING: Max disk usage exceeded but only one index directory!");
              # only one index dir we can't continue
              mame_log("WARNING:max_disk_usage exceeded but the oldest index directory is the current index.");
           }
           $warning_oversize = 1;
	   return 1; # prevent a deadlock by exiting.
       }
    } 
    return 1;
}


#=======================================================================
# update_times: reconstruct all /var/log/yyyy-mm-dd/mamelog.time(.gz)
#   at start-up. Data format is "123456789 alphats omegats".
#   for first/last times in the corresponding day's mamelog.123456789(.gz).

sub update_times {
    my $self = shift;
    my $gunzip_to_stdout  = $PRODUCT->get_util('gunzip_to_stdout');
    debug("Indexer: update_times is looking in root dir of '$self->{BASEDIR}' for directories");

    for my $dir (glob("$self->{BASEDIR}/20??-??-??")) {
        my %times;

        debug("Indexer: we are in update_times dir '$dir' for updating");
        if (open my $tfh, '<', "$dir/mamelog.time") {
            while (<$tfh>) {
                my ($key, @range) = split;
                $times{$key} = \@range;
            }
            close $tfh;
        }

        my $tfh = $self->open_out("$dir/mamelog.time");
        
                                    # NumZeros
        for my $file (glob("$dir/mamelog.?????????[0-9]{,.gz}")) {

            my ($seek) = $file =~ m{^$dir/mamelog\.(\d+)};
            next if $times{$seek};
            my %range;

            for my $bin (qw(head tail)) {

                my $bin_cmd = $PRODUCT->get_util($bin . '_n');
                my $cmd = $file =~ /\.gz$/
                            ? "$gunzip_to_stdout $file | $bin_cmd 1"
                            : "$bin_cmd 1 $file";
                $range{$bin} = substr(qx($cmd), 0, 19);
                debug("Indexer: unzip doing it a $cmd");

            }

            print $tfh "$seek $range{head} $range{tail}\n";
        }

        close $tfh;

    }
    return;
}


sub _indexer_data {
    my $self = shift;
    return $mamelog;
}

sub _should_fingerprint_email {
    my $self = shift;
    my $trans = shift;

    return 1 if ($self->{MONITOR_LOCAL_TRAFFIC});
    return 0 if $trans->field('RELAY') eq 'local';

    my $addr = $trans->field('TO');
    return 1 if ($addr !~ /\@(?:$hostname|$fqdomain|localhost)$/);
    return 0;
}

# XXX: TODO:  This is also in MTA/Postfix
sub is_blank {
    my ($self,$val) = @_;
    return (
       not $val
        or $val eq '-'
        or (ref $val eq 'HASH' and not %$val)
        or (ref $val eq 'ARRAY' and not @$val)
    ) ? 1 : 0;
}

sub _rename_and_delete_dir {
    my $dir = shift;
    # Rename the directory so that it can't be searched while we're doing
    # the rmtree
    my $renamed_dir = dirname($dir) . 'tobedeleted' . basename($dir);
    rename $dir, $renamed_dir;
    unless ( eval { rmtree($renamed_dir) } ) {
        mame_log("Unable to remove expired directory '$dir': $@");
        exit 0;
    }
    return 1;
}

# ===============
# Search-related functions


# Return the column number for a given field in the main log file
sub column_number_of_field {
    my ($self, $field) = @_;
    
    my $trans = Sophos::LogSearch::Transaction->new();
    my $trans_fields = $trans->field_names;

    my $col_num;

    # XXX: This should be a *lot* more elegant
    if ($field eq 'TIME') {
        # XXX: we should just convert ATIME to TIME
        $col_num = $trans->field_index('ATIME');

    } elsif ($field eq 'MSEEK') {
        $col_num = scalar(@{$trans_fields}) + 0;

    } elsif ($field eq 'PFQID') {
        $col_num = scalar(@{$trans_fields}) + 1;

    } elsif ($field eq 'ASEEK') {
        $col_num = scalar(@{$trans_fields}) + 2;

    } elsif ($field eq 'ZSEEK') {
        $col_num = scalar(@{$trans_fields}) + 3;

    } else {
        # One of the transaction fields
        $col_num = $trans->field_index($field);
    }

    return $col_num;
}

1;

