package Sophos::LogSearch::Source::Maillog;

use strict;
use warnings;

use Carp                         qw(carp cluck croak verbose);
use Date::Parse                  qw(str2time); # XXX: Remove if UTC function below gets removed
use POSIX                        qw(strftime &WNOHANG);
use ActiveState::Handy           qw(run);
use IO::Select;
use Sophos::LogSearch::Constants qw($PRODUCT &debug $MAMELOGD_QUEUE_LIFE $CONTINUE
                                    &mame_log utc_time_string_to_epoch  $KILL_SIG local_time_string_to_epoch
                                    utc_time_string_to_epoch $P format_date &READ_RETRY
                                    );
use Sophos::LogSearch::MTA::Postfix;
use Sophos::LogSearch::Transaction;
use Sophos::LogSearch::Config;

my $maillog;
    # {file}        Linked maillog pathname; "yyyy-mm-dd/hhmmss"
    # {in}          Input handle.
    # {line}        Current line (may be undef)
    # {next_tell}   File size after last read.
    # {paths}       List of initial maillog names (next_mail consumes)
    # {queued_time} time() at last put>>@queue.
    # {tail}        Sophos::Tailer("$self->{VARLOG}/maillog")
    # {tell}        Input-file offset of current line.

my $mta = Sophos::LogSearch::MTA::Postfix->new();

# BIG FAT TIMEZONE COMMENT:  this module only every puts UTC dates into it's 'time' values.

sub new {
    my ($class, %args) = @_;

    my $self = {

        VARLOG               => $Sophos::LogSearch::Config::config{path}->{indexes},
        MAILLOG              => $Sophos::LogSearch::Config::config{path}->{'maillog'}, # $PRODUCT->get_file_location('maillog'),
        MAMELOGD_QUEUE_LIFE  => $MAMELOGD_QUEUE_LIFE,

        # Internal variables:
        'time'               => undef,  # Last valid input time; "yyyy-mm-dd hh:mm:ss".
        has_line             => 1,      # Whether the last call to next() actually got a 
                                        # line from the MTA
        tailing              => 0,      # Whether or not we are currently tailing the maillog
        line                 => undef,  # The most recent line from the tailer

        pingless_loops       => 0,      # Number of times we read from the tailer without receiving any data
        %args,

    };
    return bless $self, $class;
}


sub init {
    my ($self, $startime) = @_;
    
    my $bin = $PRODUCT->get_util('maillog_plain');

    $maillog->{tell} = $maillog->{next_tell} = 0;
    $maillog->{queued_time} = time;
}


# Bring the tailer back up to the given time, or if that time no longer exists,
# start at the oldest maillog
sub resume {
    my ($self, $time, $resume_line, $seek) = @_;

    # time: like '2007-07-03T22:27:11'
    debug("Maillog: start resume (time: '" . ($time||'') . "' / line: '". ($resume_line||'')
            . "' seek: " . ($seek || '') . ")");
    if ($seek && $CONTINUE) {    
        $self->tail_new_file($self->{MAILLOG}, $seek, $resume_line);

    } elsif ($resume_line) {
        $self->tail_new_file($self->{MAILLOG}, 0, $resume_line);

    } else {
        $self->tail_new_file($self->{MAILLOG});
    }

    my $line;
    my $have_skipped_to_end = 0;

    while (!$line && $CONTINUE) {
        $line = $self->next_tailer_line;
        select(undef,undef,undef,0.1);
    }

    if ($line =~ /-- gap/) {  # Gap

    debug("Maillog: resume: Gap found");
        $have_skipped_to_end = 1;

    } elsif ( !(($line =~ /^#/o) || ($line =~ /^(\d+:)?\s*$/o)) ) {
        # Don't lose this line - save it so we can process it later 
        debug("Maillog: resume: resume line found");
        $self->{resume_line} = $line;
    }

    $self->{tailing} = 1;
    debug("Maillog: resumed to '$line'") if $line;
    debug("Maillog: resume: done");
    return $have_skipped_to_end;
}


# Returns whether or not a line was returned by the tailer the last time
# next() was called, regardless of whether or not that function actually
# returned a transaction
sub has_line {
    my ($self) = @_;
    return $self->{has_line};
}


# Returns the last maillog line read from the tailer
sub line {
    my $self = shift;
    return $self->{line};
}


# Returns the time of the last read line in UTC Unix time format
sub time {
    my ($self) = @_;
    if (defined $self->{time}) {
        return $self->{time_epoch} if $self->{time_epoch};
        return utc_time_string_to_epoch($self->{time});

    } else {
        return 0;
    }
}


sub queued_time {
    my ($self) = @_;
    return $maillog->{queued_time};
}

sub stop {
    my ($self) = @_;

    $self->stop_maillog_daemon;
}


# Process the next mail.log line and return a transaction object filled
# with the information gathered from that line
sub process_next_line {
    my ($self, $line) = @_;

    debug("Maillog: process_next");

    if ($line) {
        debug("Maillog: process_next: doing the processing");
        
        $self->{has_line} = 1;
        my $trans = $mta->process_line($line);

        if (defined $trans) {
            $trans->add(ATIME => $self->{time});
            $trans->add(ATIME_EPOCH => $self->{time_epoch});
            $trans->add(mail_log => "$line\n");
            return $trans;
        }

    } else {  # We didn't get a line from the tailer

        debug("Maillog: process_next: no line found");
        $self->{has_line} = 0;
    }

    return undef;
}

my $MAX_LOOPS_WITHOUT_DATA = 4 * 800; # around 4 seconds on a fully loaded system

# Read the next mail.log line from the daemon
sub next_tailer_line {
    my $self = shift;
    my $line;
    $P->begin('next_tailer_line');
    debug('Maillog: Checking for log lines from our tail pipe');

    my $fh = $maillog->{fh};
    my $status = 0;
    my $count = 1;
    my $index = index($maillog->{buffer},"\n"); # -1 is not found
    if (($index == -1) and $maillog->{ioselect}->can_read(0)) {
        $count = sysread($fh,$maillog->{buffer},1024 * 2,length($maillog->{buffer}));
	    $maillog->{pingless_loops} = 0 if $count;
        $index = index($maillog->{buffer},"\n");
    }
       
    if ($index != -1) {
        $line = substr($maillog->{buffer},0,$index +1,''); # get and remove the data
        $status = 1;
        $maillog->{pingless_loops} = 0;

    } else {
        $P->end('next_tailer_line');
        $maillog->{pingless_loops}++;
        return; # no data
    } 
    chomp($line) if $line;

    $P->end('next_tailer_line') unless $status;
    return undef unless $status;
    debug('Maillog: io select said there was something');

    if($maillog->{pingless_loops} > $MAX_LOOPS_WITHOUT_DATA) {
        $maillog->{pingless_loops} = 0;
        waitpid($maillog->{pid},&WNOHANG) if $maillog->{pid};
        $maillog->{pid} = undef; 
        $self->resume($self->{time}, $self->{line},$maillog->{mail_seek});
    } else {
       # $maillog->{pingless_loops}++;
    }

    debug("Maillog: tail pipe returned: '$line'") if $line;
    debug("Maillog: tail pipe returned NOTHING") unless $line;
    $P->end('next_tailer_line');
    return $line;
}


# Start tailing the latest maillog file from the end
sub tail_latest_file {
    my $self = shift;
    $self->tail_new_file($self->{MAILLOG});
}


# Start tailing the given file
sub tail_new_file {
    my ($self,$file,$offset,$starts_with) = @_;

    my $bin = $PRODUCT->get_util('maillog_plain');
    die "Unable to determine the location of utility 'maillog_plain'" unless $bin; 
   
    my $cmd = "$bin --filename=$file";

    if (defined($offset) and defined($starts_with)) {
        # XXX TODO escape shell quotes
        $cmd .= " --offset=$offset --starts_with='$starts_with'"
    } 
    debug("Maillog: maillog resume command: $cmd");

    if ($maillog->{pid}) {
      debug("Have a tailer running at $maillog->{pid}, sending it a $KILL_SIG and closing handle");
      
      close $maillog->{fh};
      $maillog->{fh} = undef;
      
      my $kid = undef;
      do {
          $kid = waitpid($maillog->{pid}, &WNOHANG);
          sleep 1 unless $kid == -1 or $kid == $maillog->{pid};
      } until $kid == -1 or $kid == $maillog->{pid};

    }

    my $fh;
    my ($read,$write);
    pipe($read,$write);
    if (my $pid = fork) {
        # Parent
        $maillog->{pid} = $pid;
        debug("maillog pid is $pid");
        close $write;
    } else {
        # Child
        open STDOUT, ">&=" . fileno($write);
        close $read;
        close $write;
        exec "$cmd";
        die "Failed to exec $bin: $!";
    }

    my $io = IO::Select->new();
    $io->add($read);
    $maillog->{log_filename} = $file;
    $maillog->{fh} = $read;
    $maillog->{ioselect} = $io;
    $maillog->{buffer} = '';
    $maillog->{pingless_loops} = 0;
    delete($maillog->{mail_seek});
    $maillog->{last_ping} = CORE::time;
    $self->{tailing} = 1;
    return;
}


# Get the next line from the tailer and return a Transaction object
# containing the information gathered from that line
sub next {
    my $self = shift;
    $P->begin('next_maillog_line'); 

    my $line = $self->next_maillog_line;
    $P->end('next_maillog_line'); 
    $P->begin('process_next_line'); 
    my $trans = $self->process_next_line($line);
    $P->end('process_next_line'); 
    return $trans;
}

sub statistics {
  my $self = shift;
  return ( lines => $maillog->{count} );
}


sub tailing {
    my ($self, $val) = @_;
    $self->{tailing} = $val if defined $val;
    return $self->{tailing};
}


# Get the next maillog line from the tailer and return it
sub next_maillog_line {
    my $self = shift;

    # Repeat until (line,time) are valid or (tail) returns undef:
    my $line;

    while ($CONTINUE) {

        if ($self->{resume_line}) {

            $line = $self->{resume_line};
            delete $self->{resume_line};

        } elsif ($self->{tailing}) {
            debug("Maillog: Next tail line");

            # our tests don't run forever, they only run once and determine that they are done reading the logfile 
            # because they no longer read data from it... only because of the nature of our tailer that is not
            # true all the time, this adds a loop to check our tail pipe 3 times waiting a second between checks
            # before saying we are done with the log file (at least when running tests). 
            $P->begin('next_maillog_line: line validation');

            if (READ_RETRY) {
                my $count = 0;
                $line = $self->next_tailer_line;
                RETRY_LOOP: while((!defined $line)|| ($line =~ /^(\d+:)?\s*$/o) || ($line =~ /^#/o)) {
                  $count++;
                  if($count > 3){
                    $line = undef;
                    last RETRY_LOOP;
                  }
                  select(undef,undef,undef,0.1);
                 $line = $self->next_tailer_line;
                }
            } else {
                $line = $self->next_tailer_line;
                if (defined($line)) {
                    if (($line =~ /^(\d+:)?\s*$/o) || ($line =~ /^#/o) ) {
                        debug("Maillog: undeffing line");
                        $line = undef;
                    } else {
                        debug("Maillog: actually got a line");
                    }
                }
            }

	        $P->end('next_maillog_line: line validation');
            
            return undef if !$line;

        } else {
            
            # if we don't have a {tail} pipe then open one up now, this is used to bootstrap the loop

            # Otherwise, start tailling the active mail log

            mame_log("Tailing $self->{MAILLOG}");
            debug("Maillog: Tailing $self->{MAILLOG}");

            $self->tail_new_file($self->{MAILLOG});
            $self->{tailing} = 1;

            $maillog->{tell} = $maillog->{next_tell} = 0; # reset our counters
            next;
        }

        if ($line) {
            if ($line =~ /^([0-9]+):(.+)$/o ) {
                $maillog->{mail_seek} = $1;
                $line = $2;
                $maillog->{count}++;
            } elsif($line =~ /^--/) {
                # save this data to our file.. 
                mame_log("Gap when resuming mail file.  Some data may not be indexed");
                next;
            } else {
                #warn  "Odd we recieved a line we don't recognize: '$line'\n";
                next;
            }

            $self->{line} = $line;

            my $TEMP_TIME = substr($line,0,15);
            # XXX TODO: validate the format of TEMP_TIME HERE
	    $P->begin('Maillog local_time_string_to_epoch');
            my $epoch = local_time_string_to_epoch($TEMP_TIME);
	    $P->end('Maillog local_time_string_to_epoch');
	    $P->begin('maillog format_date');
	    $self->{time} = format_date(gmtime($epoch));
            $self->{time_epoch} = $epoch;
	    $P->end('maillog format_date');
            last if $self->{time};

        } else {
            die "next_maillog_line: no line found";
        }

    }
    
    return $line;
}

sub mail_seek {
    return $maillog->{mail_seek};
}

# Stop the maillog tailing daemon
sub stop_maillog_daemon {
    my ($self) = @_;
    close $maillog->{fh} if $maillog->{fh};
    close $maillog->{tee_fh} if $maillog->{tee_fh};
    debug('Stopping maillog daemon');
    kill $KILL_SIG,$maillog->{pid} if $maillog->{pid};

    return;
}

1;
