package Sophos::LogSearch::Source::MessageLog;

use strict;
use warnings;

use Date::Parse                  qw(str2time); # XXX: Remove if UTC function below gets removed
use POSIX                        qw(strftime _exit);

use PureMessage::MessageLog         qw(split_line);
use Sophos::LogSearch::Constants qw($PREFIX $PRODUCT $DONE_MESS &debug $CONTINUE
                                    $MAMELOGD_QUEUE_LIFE &mame_log local_time_string_to_epoch
                                    utc_time_string_to_epoch $P format_date &READ_RETRY
                                    );
use Sophos::LogSearch::Transaction;
use Sophos::LogSearch::Config;
use Sophos::LogSearch::LogTail;

my $messlog;
    # {in}          Input handle
    # {line}        Current line (may be undef)
    # {paths}       List of initial message_log names (next_mail consumes)
    # {tail}        Sophos::Tailer("$self->{PMXLOG}/message_log")
    # {seek}        Byte offset into messagelog for the beginning of the current line

# THE ONLY TIMEZONE COMMENT YOU'LL EVER NEED: this module only keeps 'time' as UTC all times are translated when read
sub new {
    my ($class, %args) = @_;

    my $self = {

        MESSLOG     => $Sophos::LogSearch::Config::config{path}->{messagelog}, # $PRODUCT->get_file_location('messagelog'),
        MAMELOGD_QUEUE_LIFE  => $MAMELOGD_QUEUE_LIFE,

        'time' => undef,  # Last valid input time "yyyy-mm-dd hh:mm:ss"
        has_line => 1,    # Whether we got a line from the tailer on the last 
                          # call to next()
        
        resume_line_present => undef,  # Flag indicating whether there is a 
                                       # line left over from resuming
        %args,

    };
    return bless $self, $class;
}

sub resume {
    my ($self, $time, $resume_line, $seek) = @_;

    debug("Messagelog: start resume");

    if (defined $seek) {
        debug("Messagelog: resume: resuming at '$seek'");
        $messlog->{tail} = Sophos::LogSearch::LogTail->new(
            filename => $self->{MESSLOG}, 
            return_tell => 1,
            offset => $seek,
            starts_with => $resume_line,
	    );

    } else {
        # No resuming to do, so just start up the tailer
        debug("Messagelog: resume: tailing latest");
        $self->tail_latest_file;
    }
    
    my $line;
    my $have_skipped_to_end = 0;

    my $retries = 0;
    while (!$line && ($retries < 3) && $CONTINUE) {
        debug("Messagelog: resume: reading from tailer");
        $line = $messlog->{tail}->read;
        select(undef,undef,undef,0.1);
        $retries++;
    }

    if (defined $line) {
        if ($line =~ /-- gap --/) {  # Gap

            debug("Messagelog: resume: gap found!");
            $have_skipped_to_end = 1;

        } elsif ( !(($line =~ /^#/o) || ($line =~ /^(\d+:)?\s*$/o)) ) {
            # Don't lose this line - save it so the 
            debug("Messagelog: resume: resume line found!");
            $self->{resume_line} = $line;
        }
    }
    debug("Messagelog: resumed to line '$line'") if $line;
    debug("Messagelog: end resume");
    return $have_skipped_to_end;
}


# Returns whether or not the last call to next() returned a line
sub has_line {
    my ($self) = @_;
    return $self->{has_line};
}


# Returns the last line read from the tailer
sub line {
    my ($self) = @_;
    return $messlog->{line};
}


# Return the current seek into the messagelog
sub seek {
    my $self = shift;
    return $messlog->{seek} || 0;
}


# Returns the time of the last line from the tailer, in Unix UTC format
sub time {
    my ($self) = @_;
    if (defined $self->{time}) {
        return $self->{time_epoch};
        #return utc_time_string_to_epoch($self->{time});
       
    } else {
       return 0;
    }
}


# Get the next line from the tailer, and return
sub next {
    my $self = shift;

    my $line;
    $P->begin('messagelog source next_messagelog_line');
    $line = $self->next_messagelog_line;
    $P->end('messagelog source next_messagelog_line');

    if ($line) {
        $P->begin('messagelog source process_next_line');
        my $trans = $self->process_next_line($line);
        $trans->add(ETIME => $self->{time});
        $trans->add(ETIME_EPOCH => $self->{time_epoch});
        $self->{has_line} = 1;
        $P->end('messagelog source process_next_line');
        $messlog->{count}++;
        return $trans;

    } else {  # We haven't gotten a line from the tailer

        $self->{has_line} = 0;
        return undef;
    }
}

sub statistics {
  my $self = shift;
  return ( lines => $messlog->{count} );
}

# Process a line and return a transaction object with the information
# gathered from that line
sub process_next_line {
    my ($self, $line) = @_;

    debug("Processing message_log line");
    my ($ts, @info) = split_line($line);
    
    my (%message_info, %info);
    
    # NOTE keys have been lowercased
    while( @info ) {
        my ($k,$v) = ( shift(@info),shift(@info) ); 
        $k = lc $k;
      
        # these will get populated by pmx_action setting below 
        next if $k eq 'action';
        next if $k eq 'reason';

        # mamelog can't cope with spaces in values, so transcode them all to _
        # for subject key 's':
        if (defined($v) and ($k eq 's')) {
          $v =~ s/_/__/g;
          $v =~ s/ /_/g;
        }

        $info{$k} = [$info{$k}] if exists($info{$k}) and !ref($info{$k});
        if (ref($info{$k}) eq 'ARRAY') {
            push(@{$info{$k}}, $v);
        } else {
            $info{$k} = $v;
        }
    }

    $message_info{milter_id} = $info{q};
    $message_info{FUR} = $info{fur};
    $message_info{SUBJECT} = $info{s};
    $message_info{FROM} = $info{f};
    $message_info{FROM} =~ s/<(.*)>/$1/sg;

    # consume pmx_action lines from the milter
    my @actions;
    if ($info{'pmx_action'}) {
        @actions = ref($info{'pmx_action'}) eq 'ARRAY' ?  @{ $info{'pmx_action'} } : $info{'pmx_action'};

    } else {
        mame_log("Messagelog is in the wrong format, lacks pmx_action attributes");
        @actions = qw(Unknown);
    }

#    my $q = $info{q};
#    $messtab->{$q}{FUR} = $info{fur};

    # Deal with messages routed or copied using pmx_route.  There is no extra
    # index for these, so add in some extra info to the downstream relay
    # field.
    my $is_copy = 0;
    if (exists $info{route}) {
        $info{route} =~ s/,/~/g;
        push @{$message_info{RELAY}}, "route=$info{route}";
        # The 'route' action will show up in the pmx_action mark, so don't do
        # anything here
    }
    if (exists $info{copy}) {
        $info{copy} =~ s/,/~/g;
        push @{$message_info{RELAY}}, "copy=$info{copy}";
        # The 'copy' action won't show up in pmx_action (as it's not the final
        # action for the message, so we need to add it in manually.
        $is_copy = 1;
    }

    foreach my $store_key ( qw(action reason to groups aliases) ) {
        # promote these keys all to array refs
        $message_info{uc($store_key) } = [];
    }
    
    my $pmx_action = '';

    # Parse the pmx_action key
    foreach my $action ( @actions ) {
        # mamelog can't cope with spaces, so encode them if they exist
        $action =~ s/_/__/g;
        $action =~ s/ /_/g;

        my @data = split(',',$action);
        my %data = map { ($_ => shift(@data) ) } qw(action reason groups to);
        
        $data{aliases} = shift @data;    # XXX multiple aliases will only result in the first being added
          
        foreach my $store_k ( qw( action reason to groups aliases ) ) {
            if ($store_k eq 'action' && defined($data{$store_k})) {
                my %actions = ( 'keep' => 'deliver' );
                $data{$store_k} = $actions{$data{$store_k}} if $actions{$data{$store_k}};
            }
            push(@{$message_info{ uc($store_k) }}, $data{$store_k}) if defined($data{$store_k});
        }

    }

    push @{$message_info{ACTION}}, 'copy' if $is_copy;

    # Now convert these arrays to strings
    foreach my $store_k ( qw(action reason to groups aliases relay) ) {
        $message_info{uc($store_k)} = join(',', @{ $message_info{ uc($store_k) } } )
                if exists $message_info{uc($store_k)};
    }
    
    foreach my $unsafe_key ( qw(SUBJECT FROM TO ALIASES) ) {
         $message_info{$unsafe_key} =~ tr/\x00-\x1f//d if exists($message_info{$unsafe_key}) and defined($message_info{$unsafe_key});
    }


    $message_info{done_message_log} = 1;
    my $trans = Sophos::LogSearch::Transaction->new(%message_info);

    return $trans;
}

# read next line from message_log*, setting (line,time).
sub next_messagelog_line {
    my $self = shift;

    my $line;

    # Repeat until (line,time) are valid, or (tail) returns undef:
    while ($CONTINUE) {

        if (defined $self->{resume_line}) {

            $line = $self->{resume_line};
            delete $self->{resume_line};

        } elsif ($messlog->{tail}) {
            
            $P->begin('messlog_tail');
            if (READ_RETRY) { 
                my $count = 0;
                $line = $messlog->{tail}->read; # read(nowait) may return ()
                RETRY_LOOP: while((!defined $line) || ($line =~ /^#/o) || ($line =~ /^(\d+:)?\s*$/o) ) {
                  $count++;
                  if($count > 3){
                    $line = undef;
                    last RETRY_LOOP;
                  }
                  select(undef,undef,undef,0.1);
                  $line = $messlog->{tail}->read; # read(nowait) may return ()
                }
            } else {

                $line = $messlog->{tail}->read; # read(nowait) may return ()
            }
            $P->end('messlog_tail');
            
            return undef if !defined $line;
        } else {
            $self->tail_latest_file;
            next;
        }

        chomp $line;

	if (!$line || ($line =~ /^--/o) || ($line =~ /^#/o) || ($line =~ /^(\d+:)?\s*$/o)) {
	   return;
        } elsif ($line =~ /^([0-9]+):(.+)$/o ) {
           $messlog->{seek} = $1;
           $line = $2;
        }
        $self->{time} = mess_timestamp($line);
        my $result = $self->{time_epoch} = utc_time_string_to_epoch($self->{time});
        last if $result;
    }
    $messlog->{line} = $line;
    return $line;
}


# Start tailing the latest message_log file at the very end
sub tail_latest_file {
    my $self = shift;

    # If we're already tailing, stop the existing tailer
    if (defined $messlog->{tail}) {
	debug("Messagelog: tail_new_file: stopping old tailer");
        $messlog->{tail}->stop;
        $messlog->{tail} = undef;
    }
    # Default behaviour: seek to the end of the file and tail from there
	debug("Messagelog: tail_new_file: tailing '$self->{MESSLOG}'");
	$messlog->{tail} = Sophos::LogSearch::LogTail->new(
		filename => $self->{MESSLOG}, 
		return_tell => 1,
		);
}


# convert all non UTC times to UTC
sub mess_timestamp {
    my ($line) = @_;
    my ($timestamp, $zone) = $line =~ /^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})(Z?)/;
    unless ($zone) {
        if ($line =~ qr/^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})/) {
	        my $epoch = local_time_string_to_epoch($1);
		$timestamp = format_date(gmtime($epoch));
	    }
    }
    return $timestamp;
}


sub init {
    my ($self,$startime) = @_;
    my $tail_n = $PRODUCT->get_util('tail_n');
    my @messlogs = reverse grep { /\.\d+$/ } sort glob "$self->{MESSLOG}.[0-9]*[0-9]";

    @messlogs = grep {substr(qx($tail_n 1 $_),0,19) ge $startime} @messlogs if $startime;
    $messlog->{paths} = \@messlogs;
    return;
}

1;

