package Sophos::LogSearch::Transaction;

use strict;
use warnings;

use Sophos::LogSearch::Constants qw($P utc_time_string_to_epoch);

# ===============
# Constants
 
# Ordered list of the fields to be written 
our @FIELDS = qw(ATIME ACTION FROM MSGID RELAY CLIENT TO REASON SUBJECT ALIASES GROUPS);
our %FIELD_INDEX = map {$FIELDS[$_] => $_} 0..$#FIELDS;

# Data to index on (data that can be quickly searched on)
our @INDEXES = qw(from to relay groups msgid aliases mailroute);

# ===============

sub new {
    my ($class, %args) = @_;

    my $self = {};
    bless $self, $class;

    $self->{contents} = {
        %args,
    };

    return $self;
}

# ===============
# ID functions

sub id {
    my $self = shift;
    return $self->{contents}->{id};
}

sub milter_id {
    my $self = shift;
    return $self->{contents}->{milter_id};
}

sub id2 {
    my $self = shift;
    return $self->{contents}->{id2};
}

sub done_message_log {
    my $self = shift;
    return $self->{contents}->{done_message_log};
} 

sub done_mail_log {
    my $self = shift;
    return $self->{contents}->{done_mail_log};

    # If we have a second id, the mail log is not complete until we have 2 removed lines
    if (exists $self->{contents}->{id} and exists $self->{contents}->{id2}) {
      return (exists $self->{contents}->{done_mail_log} and ($self->{contents}->{done_mail_log} > 1));
    } else {
      return $self->{contents}->{done_mail_log};
    }
} 

sub done {
    my $self = shift;        

    if (exists $self->{contents}->{id} and exists $self->{contents}->{id2}) {
       # we need TWO removed lines
       return (exists($self->{contents}->{done_mail_log}) and ($self->{contents}->{done_mail_log} > 1) && 
                $self->{contents}->{done_message_log});

    } else {
        return ($self->{contents}->{done_mail_log} && 
                $self->{contents}->{done_message_log});
    }
}

# ===============
# Content-modifying functions

sub contents {
    my $self = shift;
    return $self->{contents};
}

sub set_contents {
    my ($self, $new_contents) = @_;
    $self->{contents} = $new_contents;
}

sub add {
    my ($self, $key, $val) = @_;
    $self->{contents}->{$key} = $val;
}


# ===============
# Index creation functions

sub field_names {
    my $self = shift;
    return \@FIELDS;
}


sub index_names {
    my $self = shift;
    return \@INDEXES;
}


# Get a named field
sub field {
    my ($self, $field) = @_;

    my $field_data = $self->{contents}->{$field} || '-';
    # Protect ourselves against invalid fields that have spaces
    if ( ($field eq 'FROM') || ($field eq 'MSGID') ) {
        $field_data =~ tr/ /_/;
    } elsif ( ref($field_data) eq 'HASH' ) {
        $field_data = join(',', keys(%{$field_data}));
    }

    return $field_data;
}

# Returns the field data to be written by indexer
sub get_fields {
    my $self = shift;

    my $output_fields = [];

    foreach my $field (@FIELDS) {
        push @{$output_fields}, $self->field($field);
    }

    return $output_fields;
}


# Returns the index number of the field in @FIELDS
sub field_index {
    my ($self, $field) = @_;
    return $FIELD_INDEX{$field};
}


# Get all data to write to indexes
sub get_indexes {
    my $self = shift;

    my $index_data = {};

    foreach my $index (@INDEXES) {

        my $val = $self->field( uc($index) );
        if (($val !~ /^-!/) && ($val ne '-')) {
            $index_data->{$index} = $val;
        }
    }

    # 'TO' should also be searchable in aliases:
    if ( exists $index_data->{to} ) {

        $index_data->{aliases} .= ',' if exists $index_data->{aliases};
        $index_data->{aliases} .= $index_data->{to};
    }

    # Mailroute includes:
    #  - FUR
    #  - Relay
    #  - Client
    #  - Milter host 
    my $fur = $self->field('FUR');
    if ($fur ne '-') {
        $index_data->{mailroute} .= ',' if exists $index_data->{mailroute};
        $index_data->{mailroute} .= $fur;
    }

    return $index_data;
}


# ===============
# Other data return functions 

sub time {
    my $self = shift;
    my $time = undef;

    if ( exists($self->{contents}->{ATIME_EPOCH}) ) {
        $time = $self->{contents}->{ATIME_EPOCH};
    
    } elsif ( exists($self->{contents}->{ETIME_EPOCH})) {
        $time = $self->{contents}->{ETIME_EPOCH};

    } elsif(exists $self->{contents}->{ATIME}) {
        # Has seen a mail.log
        $time = utc_time_string_to_epoch($self->{contents}->{ATIME});


    } elsif (exists $self->{contents}->{ETIME}) {
        # Has seen a message_log only
        $time =  utc_time_string_to_epoch($self->{contents}->{ETIME});
    }
    return $time if $time;
    return;
}

1;
