package ActiveState::Config;

use strict;
use base qw(Exporter);

our $VERSION = '5.500';
our @EXPORT_OK = (qw(
    path_nodes
    path_node
    path_value
    path_values
    path_value_append
    path_block_comment
    path_line_comment
    path_children
));
our %EXPORT_TAGS = (path => \@EXPORT_OK);

# Prototyped subs
sub vdie ($$) {
    my ($toke, $msg) = @_;
    $msg =~ s/\s+\z//;
    my $loc = '';
    if (defined $toke and UNIVERSAL::isa($toke, "ActiveState::Config::Tokenizer")) {
	my $file = $toke->filename;
	my $line = $toke->lineno;
	$loc = "$file:$line: ";
    }
    die "$loc$msg\n";
}

# Joining and splitting paths:
sub group { join '#', grep { defined $_  && length $_ } @_ }
sub groupsplit { my @a = split /#/, $_[0]; @a == 2 ? @a : () }
sub path { join '.', grep { defined $_ && length $_ } @_ }
sub pathsplit { split /\./, $_[0] }

sub new {
    my $cls = shift;
    my $o = bless {
	opts => {
	    allow_includes => 1,	# allow the !include directive
	    strict => 0,		# allow un-typedef'd nodes
	    # an arrayref. The first element is the default #type for all
	    # nodes. The subsequent elements are #types which, if seen, will
	    # avoid using the default. So if you pass ['#atom', '#list'], then
	    # all nodes will automatically be declared as #atom, even if you
	    # give them #float type. The only exception is if you declare a
	    # primary type as a #list -- this avoids #atom.
	    default_type => undef,
	    # if you want backup files created when you call the save()
	    # method, specify what extension they should have. Currently there
	    # is no "logrotate-like" facility for keeping N levels of backups.
	    backup_ext => undef,
	    is_utf8 => undef, #allow string data to remain in utf8. If not set
	    # we will force conversion to latin1.
	    @_,
	},
    }, ref($cls) || $cls;
    $o->{__tree__} = $o->node(
	type => 'group',
	path => '',
	groupclass => '',
    );
    $o->ActiveState::Config::Types::register_types;
    {
	package ActiveState::Config::FakeObject;

	# Breaks circular references in the parser, freeing memory
	sub DESTROY {
	    my $fake = shift;
	    my $p = $$fake;
	    delete $p->{__tree__};
	    delete $p->{tokenizers};
	    delete $p->{toke};
	}
	sub AUTOLOAD {
	    my $fake = shift;
	    my $o = $$fake;
	    our $AUTOLOAD;
	    (my $method = $AUTOLOAD) =~ s/^.*:://;
	    return $o->$method(@_);
	}
    }
    bless \$o, "ActiveState::Config::FakeObject";
}

# This is the user API for parsing a file at the toplevel. Internally, it just
# calls _parse_file() with the root path ('').
sub parse_file {
    my ($o, $file, @opts) = @_;
    push @opts, is_utf8 => $o->{opts}{is_utf8};
    $o->_parse_file($file, '', backup_ext => $o->{opts}{backup_ext}, @opts);
}

sub parse {
    my ($o, $file, @opts) = @_;
    my $toke = $o->_tokenizer(buffer => $file, @opts);
    $o->_parse($toke, '', @opts); # always parses from the toplevel
}

sub root {
    my $o = shift;
    $o->{__tree__}->autovivify(1);
    $o->{__tree__};
}

# Return the serialized version
sub as_string {
    my $o = shift;
    $o->{__tree__}->_as_string;
}

# Save to a specific file
sub saveto {
    my ($o, $file) = @_;
    $o->{__tree__}->_saveto($file);
}

# Save to the original files (requires writable => 1)
sub save {
    my $o = shift;
    $o->{__tree__}->_save;
    $o->unlock;
}

# Convert the tree to a hash
sub as_hash {
    my $o = shift;
    my $href = {};
    $o->{__tree__}->autovivify(1);
    $o->{__tree__}->_native($href);
    $href = $href->{''}; # because the root node's name is "".
    wantarray ? %$href : $href;
}

sub default {
    my ($o, $path, $value) = @_;
    if (@_ == 3) {
	$o->{default}{$path} = $value;
	return;
    }
    return $o->{default}{$path} if exists $o->{default}{$path};
    for my $defpath (keys %{$o->{default}}) {
	next unless _match_path($defpath, $path);
	return $o->{default}{$defpath};
    }
    undef;
}

sub _defaults_in_group {
    my ($o, $path) = @_;
    my %matches;
    for my $defpath (keys %{$o->{default}}) {
	my @defpath = split /\./, $defpath;
	my $last = pop @defpath;
	next if $last =~ /\*/;
	my $path_new = path(@defpath);
	if (_match_path($path_new, $path)) {
	    my $real_path = path($path, $last);
	    $matches{$real_path} = $o->{default}{$defpath};
	}
    }
    wantarray ? %matches : \%matches;
}

# Defines and assigns type declarations. There are four ways to use it,
# resulting in reasonably hairy code :)
sub typedecl {
    my $o = shift;

    # Defining a named type:
    if (@_ >= 3 and $_[0] =~ /^#/) {
	my ($t, $v, $f, $vargs, $fargs) = @_;
	$vargs = {} unless defined $vargs;
	$fargs = {} unless defined $fargs;
	$o->{typedecl}{def}{$t} = [$v, $f, $vargs, $fargs];
	return 1;
    }
    # Retrieving an existing type definition:
    elsif (@_ == 1 and $_[0] =~ /^#/) {
	my $t = shift;
	my (@parts) = split ' ', $t;
	my $d = $o->{typedecl}{def}{$parts[0]};
	return defined $d ? @$d : ();
    }
    # Retrieve a type definition by path:
    elsif (@_ == 1) {
	my $p = shift;

    # XXX: Make it predictable which order the paths that contain wildcards
    # will be in.

    # We must put any paths that contain a wildcard at the end of the list so
    # that exact strings match first.
    my @paths = ();
    my @wildcard_paths = ();
	for my $path (keys %{ $o->{typedecl}{path} }) {
        if($path =~ /\*$/) {
            push(@wildcard_paths, $path);
        }
        else {
            push(@paths, $path);
        }
    }
    @wildcard_paths = sort { length($a) <=> length($b) } @wildcard_paths;
    push(@paths, @wildcard_paths);

	my $td;
	for my $path (@paths) {
	    next unless _match_path($path, $p);
	    $td = $o->{typedecl}{path}{$path};
	    return wantarray ? @$td : $td;
	}
	return;
    }
    # Declaring a node to be a particular named type:
    elsif (@_ == 2 and $_[1] =~ /^#/) {
	my ($path, $t) = @_;
	my ($v, $f) = $o->typedecl($t) or return;
	$o->{typedecl}{path}{$path} = [$t, $v, $f];
    }
    # Declaring a node to be an unnamed type:
    elsif (@_ == 3) {
	my ($path, $v, $f) = @_;
	$o->{typedecl}{path}{$path} = [undef, $v, $f];
    }
}

sub unlock {
    my $o = shift;
    my $tokenizers = $o->{tokenizers};
    return unless $tokenizers and @$tokenizers;
    for (@$tokenizers) {
	$_->close;
    }
}

sub _tokenizer {
    my $o = shift;
    use ActiveState::Config::Tokenizer;
    my $toke = ActiveState::Config::Tokenizer->new(@_);
    push @{$o->{tokenizers}}, $toke;
    $o->{__tree__}{tokenizer} ||= $toke;
    $toke;
}

# Returns a node whose children are the nodes in the file. This node (and its
# children) will have paths set relative to the $path passed as an argument to
# the file. By default, the $path starts at '', the root of the tree.
sub _parse_file {
    my ($o, $file, $root, @opts) = @_;
    my $toke = $o->_tokenizer(file => $file, @opts);
    $o->_parse($toke, $root, @opts);
    # each node sets the dirty flag; unset it here
    $toke->{dirty} = 0;
}

my $uniq = 0;
sub _uniqid {
    my ($n, $root) = @_;
    my $name = '!uniqid:' . ++$uniq;
    my $newroot = "$root.$name";
    $n->type('placeholder');
    $n->name($name);
    $n->path($newroot);
    $newroot;
}

sub _parse {
    my ($o, $toke, $root, @opts) = @_;

    my $reset = 1;
    my ($node, $oldroot, $touched);

    # Now start parsing:
    while (1) {
	# We must set the current tokenizer in the parser object, because
	# it may have been set by recursing into other _parse() calls.
	$o->{toke} = $toke;

	# First, get ourselves a placeholder node to hold leading comments:
	if ($reset) {
	    $node = $o->node(type => 'placeholder');
	    $node->{__emit__}{infile} = 1;
	    $reset = 0;
	    $touched = 0;
	}

	# update the value in the parse tree
	my $token = eval { $toke->next };
	vdie $toke, $@ if $@;
	last unless $token;
	my $type = $token->{type};
	if ($type eq 'NAMEVALUE') {
	    $oldroot = path($root, $token->{name});
	    $node->type('option');
	    $node->name($token->{name});
	    if ($token->{special}) {
		my $val = $token->{value};
		if ($val eq '__DEFAULT__') {
		    # currently nothing
		}
		else {
		    vdie $toke, "unknown SPECIAL token '$val'";
		}
	    }
	    else {
		$node->value($token->{value});
                $node->changed(0);
		$node->{__emit__}{vtype} = $token->{vtype};
	    }
	    $o->_update($node, $oldroot)->{tokenizer} = $toke;
	    $reset = 1;
	}
	elsif ($type eq 'VALUE') {
	    my $val = $token->{value};
	    vdie $toke, "value '$val' outside of list" unless length $root;
	    $node->block_comment(''); # ignore block comments on values
	    $node->type('option');
	    $node->value($val);
            $node->changed(0);
	    $node->{__emit__}{listtype} = 'grouped';
	    $node->{__emit__}{vtype} = $token->{vtype};
	    $o->_update($node, $root)->{tokenizer} = $toke;
	    $reset = 1;
	}
	elsif ($type eq 'BLOCK_COMMENT') {
	    my $block = $node->block_comment;
	    if (defined $block and length $block) {
		$block .= $token->{value};
	    }
	    else {
		$block = $token->{value};
	    }
	    $node->block_comment($block);
	    $touched = 1;
	}
	# Update the previous node's line comment.
	elsif ($type eq 'LINE_COMMENT') {
	    my $node = path_node($o->{__tree__}, $oldroot || $root) or next;
	    $node->line_comment($token->{value});
	}
	# Can either be a list or an unnamed group -- we don't know until we
	# recurse to the next level.
	elsif ($type eq 'OPEN') {
	    my $grouptype = $token->{group};
	    my $groupname = $token->{name};
	    # Set the node type to "group" tentatively. It might turn out to
	    # be a list.
	    $node->type('group');
	    $node->name($groupname) if $groupname;
	    $node->groupclass($grouptype);
	    $oldroot = path($root, group($grouptype, $groupname));
	    $o->_update($node, $oldroot)->{tokenizer} = $toke;
	    $o->_parse($toke, $oldroot, @opts);
	    $reset = 1;
	}
	# and break out to the previous level
	elsif ($type eq 'CLOSE') {
	    my $tag = $token->{group};
	    my $open_node = path_node($o->{__tree__}, $root)
		or vdie $toke, "close tag '$tag' outside of group\n";
	    my $open;
	    if ($open_node->{type} eq 'group') {
		$open = $open_node->groupclass;
	    }
	    else {
		$open = $open_node->{name};
	    }
	    $open_node->autovivify;
	    vdie $toke, "close tag '$tag' doesn't match open tag '$open'"
		if $open ne $tag;
	    last; # we recurse into OPEN tokens, so break out upon CLOSE.
	}
	elsif ($type eq 'VERSION') {
	    my $ver = $token->{value};
	    vdie $toke, "config file version too high ($ver). Maximum $VERSION"
		if $ver > $VERSION;
	}
	elsif ($type eq 'DIRECTIVE') {
	    my ($name, $arg) = ($token->{name}, $token->{value});

	    # insert a placeholder for the emitter
	    my $tmp = _uniqid($node, $root);
	    $node->{__emit__}{directive} = "!$token->{name} $token->{value}";
	    $o->_update($node, $tmp)->{tokenizer} = $toke;
	    $reset = 1;

	    if ($name eq 'include') {
		next unless $o->{opts}{allow_includes};
		my $glob = $token->{value};
		# if $glob is a relative path, make it relative to the current
		# file.
		$glob = $toke->dir . "/$glob" unless $glob =~ m#^/#;
		use File::Glob qw(:glob);
		my @files = bsd_glob($glob);
		$o->_parse_file($_, $root, @opts) for @files;
	    }
	    else {
		vdie $toke, "unknown parser directive '!$name'";
	    }
	}
	else {
	    vdie $toke, "internal error: unknown token '$type'\n";
	}
    }
    if ($touched) {
	$oldroot = _uniqid($node, $root);
	$node->{__emit__}{block_comment_only} = 1;
	$o->_update($node, $oldroot);
    }
}

sub _update {
    my ($o, $node, $path) = @_;
    my $toke = $node->{tokenizer};
    my $prev = $o->node(path => $path, create => 0);
    unless ($prev) {
	my $n = eval { path_node($o->{__tree__}, $path, insert => $node) };
	vdie $toke, $@ if $@;
	return $n;
    }

    # Update the tokenizer:
    $prev->{tokenizer} = $node->{tokenizer};

    # If the node was not seen in the real parse stream yet (i.e. it was
    # autovivified), reinsert it into the tree here:
    unless ($prev->{__emit__}{infile}) {
	path_node($o->{__tree__}, $path, delete => 1);
	my $n = eval { path_node($o->{__tree__}, $path, insert => $node) };
	vdie $toke, $@ if $@;
	return $n;
    }

    # Change placeholder nodes into real nodes:
    my $oldtype = $prev->{type};
    if ($oldtype eq 'placeholder') {
	$prev->type($node->{type});
    }

    # Update the value in the node
    if (grep { $_ eq $oldtype } qw(option placeholder)) {
	if (defined $node->value) {
	    if ($oldtype eq 'option') {
		eval { $prev->value_append($node->value) };
		vdie $toke, $@ if $@;
		$prev->{__emit__}{listtype} = $node->{__emit__}{listtype} || 'repeated';
	    }
	    else {
		eval { $prev->value($node->value) };
		vdie $toke, $@ if $@;
	    }
	}
    }
    elsif ($oldtype eq 'group') {
	if ($node->{type} eq 'option' and $prev->children->size) {
	    my $val = $node->value;
	    vdie $toke, "can't add list entry '$val' to group '$path'";
	}
	elsif ($node->{type} eq 'option') {
	    $prev->type('option');
	    # Drop line comments into the block-comment area.
	    if ($prev->{line_comment}) {
		my $bcom = "$prev->{block_comment}$prev->{line_comment}\n";
		$prev->block_comment($bcom);
		$prev->line_comment('');
	    }
	    eval { $prev->value([$node->value]) };
	    vdie $toke, $@ if $@;
	}
	elsif ($node->{type} eq 'group') {
	    # don't do anything!
	}
    }
    $prev->block_comment($_) if defined($_ = $node->block_comment) and length $_;
    $prev;
}

# Node factory
sub node {
    my ($o, %args) = @_;
    my %node = (
	type => 'placeholder',
	name => '',
	block_comment => '',
	parser => $o,
	tokenizer => $o->{toke},
    );
    while (my ($k, $v) = each %args) {
	if (grep { $k eq $_ } qw(type name groupclass typedecl block_comment
	    line_comment value default path tokenizer)) {
	    $node{$k} = $v;
	}

	# Undocumented:
	elsif ($k eq 'create' and !$v) {
	    return undef unless my $path = delete $args{path};
	    my $cursor = path_nodes($o->{__tree__}, $path, defaults => 0);
	    return undef unless $cursor->size;
	    return $cursor->next;
	}
#	else {
#	   die "Unknown argument '$k' to node() method";
#	}
    }
    bless \%node, "ActiveState::Config::Node";
}

sub iter {
    my $o = shift;
    my $ch = @_ ? shift : [];
    $ch = [$ch] unless ref($ch) eq 'ARRAY';
    return bless [-1, $ch], "ActiveState::Config::Iter";
}

#=============================================================================
# The CPATH functions and support functions
#=============================================================================

sub _match_path {
    my ($path, $fullpath) = @_;
    if (index($path, "*") != -1) {
	$path = quotemeta($path);
	$path =~ s/\\\*/[^#.]*/g;
	return $fullpath =~ /^$path\z/;
    }
    $path eq $fullpath;
}

# Attach a node to the given path:
sub _attach {
    my ($root, $path, $node) = @_;
    return $_[0] = $node if $path eq '';

    my @segments = split /\./, $path;
    my $currentpath = $root->path;
    while (@segments) {
	my $seg = shift @segments;
	$currentpath = path($currentpath, $seg);
	my @a = split /#/, $seg;
	my ($class, $name) = @a == 2 ? @a : ();
	my $cursor = path_nodes($root, $seg, defaults => 0);
	return if $cursor->size > 1; # they used wildcards (stupid)
	$root->{parser}{toke} = $root->{tokenizer};

	# Need to attach a node to the root. If there are segments left,
	# attach a group node; otherwise attach the given node.
	if ($cursor->size == 0) {
	    my $child = $root->children;
	    if (@segments) {
		$root->type('group');
		$root = $root->{parser}->node(
		    type => 'group',
		    $name
			? (groupclass => $class, name => $name) 
			: (groupclass => $seg),
		);
		$root->path($currentpath);
		$child->append($root);
	    }
	    else {
		my $p = $node->{parser};
		my $temp = $p->node(type => $node->{type});
		$temp->path($currentpath);
		if ($temp->{type} eq 'group') {
		    $temp->name($name) if $name;
		    $temp->groupclass($class ? $class : $seg);
		}
		else {
		    $temp->name($seg);
		}
		# Assign a typedecl if one exists:
		if (my $td = $p->typedecl($currentpath)) {
		    if (defined $td->[0]) {
			$temp->_typedecl($td->[0]);
		    }
		    else {
			$temp->_typedecl(@$td[1 .. $#$td]);
		    }
		}
		# check if the parent's type is "#any". This is a special
		# case: the parser won't enforce any kind of typing on
		# anything below an #any node. All subnodes of an #any node
		# are also typed as #any.
		elsif ( ($root->typedecl || '') eq '#any') {
		    $temp->_typedecl('#any', 1);
		}
		elsif ($p->{opts}{strict} and $node->{type} eq 'option') {
		    die "Unexpected option '$currentpath' while parsing in strict mode.\n";
		}

		# Need to append to the list *before* calling absorb(). Absorb
		# ends up calling out to the validator, which may autovivify
		# the parent. That ends up adding this node so we end up with
		# multiple copies of it.
		$child->append($temp);

		# Now assign the rest of the values
		$temp->absorb($node, $p->default($currentpath));
		$temp->{tokenizer} = $root->{tokenizer};
		return $temp;
	    }
	}
	# Make sure the single matching node is a group node.
	elsif (@segments) {
	    my $n = $cursor->next;
	    my $t = $n->{type};
	    if ($t eq 'placeholder') {
		$n->type('group');
	    }
	    elsif ($t ne 'group') {
		my $t = $node->{type};
		die "can't coerce node '$currentpath' (type $t) into group.\n";
	    }
	    $root = $n;
	}
	else {
	    # Never a good idea to replace: force user to call delete first:
	    return $cursor->next; # return the node we didn't replace.
	}
    }

    $node;
}

# Returns a set of nodes that match the $path:
sub path_nodes {
    my ($root, $path) = (shift, shift);
    my %opts = (
	defaults => 1,
	@_,
    );
    my $want = wantarray;
    my $ret = sub { return $want ? @_ : ActiveState::Config->iter(\@_) };
    return &$ret() unless $root; # empty list
    return &$ret() unless defined $path;

    # Forward to _attach() if we're doing that.
    if (my $type = $opts{create}) {
	$type = 'option'
	    unless grep { $type eq $_ } qw(option group placeholder);
	my $node = $root->{parser}->node(type => $type);
	return &$ret(_attach($root, $path, $node));
    }
    elsif (my $node = $opts{insert}) {
	return &$ret(_attach($root, $path, $node));
    }
    elsif (delete $opts{delete}) {
	for my $node (path_nodes($root, $path, %opts)) {
	    my @parts = pathsplit($node->{path});
	    my $name  = pop @parts;
	    my $parent = path(@parts);
	    my $node = path_node($root, $parent, defaults => 0) or next;
	    next unless $node->{type} eq 'group';
	    my $children = $node->children;
	    while (my $child = $children->next) {
		next unless $child->{segment} eq $name;
		$children->delete;
	    }
	}
	return &$ret();
    }

    return &$ret($root) if $path eq '';

    # Have a list of "currently-selected" nodes:
    my @selected = $root;
    my @segments = split /\./, $path;
    while (@selected and @segments) {
	my $cur = shift @segments;
	@selected = grep {
	    (!@segments or $_->{type} eq 'group') and _match_path($cur, $_->{segment});
	} map { $_->children } @selected;
    }

    # If nothing was selected, check for a matching default. If one is found,
    # attach it to the tree, and return it.
    if (!@selected and $opts{defaults} and $path !~ /\*/) {
	if (my $def = $root->{parser}->default($path)) {
	    my $node = $root->{parser}->node(type => 'placeholder');
	    push @selected, _attach($root, $path, $node);
	}
    }

    &$ret(@selected);
}

sub path_node {
    my $nodes = path_nodes(@_);
    return $nodes->next;
}

sub path_value {
    my ($root, $path, $value) = @_;
    my $node = path_node(
	$root, $path,
	insert => (@_ == 3 ? ($root->parser->node(type => 'option')) : undef),
    ) or return;
    return $node->value unless @_ == 3; # set the value
    $node->value($value);
}

sub path_values {
    my ($root, $path, @values) = @_;
    my $node = path_node(
	$root, $path,
	insert => (@values ? ($root->parser->node(type => 'option')) : undef),
    ) or return;
    return $node->values unless @values; # set the values
    $node->values(@values);
}

sub path_value_append {
    my ($root, $path, $value) = @_;
    my $node = path_node(
	$root, $path,
	insert => (@_ == 3 ? ($root->parser->node(type => 'option')) : undef),
    ) or return;
    $node->value_append($value);
}

sub path_block_comment {
    my ($root, $path, $bcom) = @_;
    my $node = path_node($root, $path) or return;
    return $node->block_comment unless @_ == 3; # set block comment
    $node->block_comment($bcom);
}

sub path_line_comment {
    my ($root, $path, $lcom) = @_;
    my $node = path_node($root, $path) or return;
    return $node->line_comment unless @_ == 3; # set line comment
    $node->line_comment($lcom);
}

sub path_children {
    my ($root, $path, $children) = @_;
    my $node = path_node($root, $path) or return;
    return $node->children unless @_ == 3; # set children
    $node->children($children);
}

package ActiveState::Config::Node;

sub dirty {
    my $o = shift;
    $o->{tokenizer}{dirty} = shift if @_ and $o->{tokenizer};
    $o->{tokenizer}{dirty};
}

sub changed {
    my $o = shift;
    $o->{changed} = shift if @_;
    $o->{changed};
}

# Facility for passing hints to the emitter:
sub emit {
    my ($o, $k, $v) = @_;
    if (@_ == 3) {
	$o->{__emit__}{$k} = $v;
    }
    return $o->{__emit__}{$k};
}

sub name {
    my $o = shift;
    my $n = $o->{name};
    if (@_) {
	$o->{name} = shift;
	$o->_update_path;
	$o->dirty(1);
    }
    $n;
}

sub groupclass {
    my $o = shift;
    return undef unless $o->{type} eq 'group';
    my $cls = $o->{groupclass};
    if (@_) {
	$o->{groupclass} = shift;
	$o->_update_path;
	$o->dirty(1);
    }
    $cls;
}

sub _update_path {
    my $o = shift;
    return unless defined $o->{path}; # don't bother if path isn't defined
    # If we're passed an argument, it's an anchored substring to replace. This
    # occurs when a parent is renamed, and we have to update our path. This
    # doesn't change our name or groupclass, so we can just do a substitution:
    if (@_) {
	my ($old, $new) = @_;
	$o->{path} =~ s/^\Q$old\E/$new/;
	return $o->{path};
    }
    my $seg = '';
    if ($o->{type} eq 'group') {
	$seg = $o->{groupclass};
	$seg .= "#$o->{name}" if length($o->{name});
    }
    else {
	$seg = $o->{name};
    }
    my @parts = split /\./, $o->{path};
    $parts[-1] = $seg;
    my $oldpath = $o->{path};
    my $newpath = $o->{path} = join('.', @parts);
    $o->{segment} = $seg;

    # If we're a group node, update all of our children:
    if ($o->{type} eq 'group') {
	for my $ch ($o->children) {
	    $ch->_update_path($oldpath, $newpath);
	}
    }
    $newpath;
}

sub absorb {
    my ($o, $other, $default) = @_;
    my @fields = (qw(line_comment block_comment children));
    for my $field (@fields) {
	next unless exists $other->{$field};
	$o->{$field} = $other->{$field};
    }
    if ($other->{__emit__}) {
	for my $k (keys %{$other->{__emit__}}) {
	    $o->{__emit__}{$k} = $other->{__emit__}{$k};
	}
    }
    if (exists $other->{value} and defined $other->{value}) {
	$o->value($other->{value});
    }
    elsif (defined $default) {
	$o->_default($default);
    }
}

sub absorb_deep {
    my ($o, $other, $do_types) = @_;
    for my $field (qw(name groupclass path default)) {
	next unless exists $other->{$field};
	$o->{$field} = $other->{$field};
    }
    if ($do_types) {
	if (my $td = $o->{parser}->typedecl($o->{path})) {
	    if (defined $td->[0]) {
		$o->_typedecl($td->[0]);
	    }
	    else {
		$o->_typedecl(@$td[1 .. $#$td]);
	    }
	}
    }

    my $ch = delete $other->{children};
    $o->absorb($other);
    if (defined $ch) {
	$other->{children} = $ch;
	my @nch;
	for my $child (@$ch) {
	    my $new = $o->{parser}->node(type => $child->{type});
	    $new->absorb_deep($child);
	    push @nch, $new;
	}
	$o->{children} = \@nch;
    }
}

sub autovivify {
    my $o = shift;
    return if $o->{__autovivify__}++;
    if (defined $o->{path}) {
	my $p = $o->{parser};
	my %children = $p->_defaults_in_group($o->{path});
	for my $child (sort keys %children) {
	    my @parts = split /\./, $child;
	    my $name = pop @parts;
	    my $n = $p->node(type => 'placeholder');
	    ActiveState::Config::_attach($o, $name, $n);
	}
    }
    if (@_) {
	return unless $o->{type} eq 'group';
	for my $child (@{$o->{children}}) {
	    next if UNIVERSAL::isa($child->{tokenizer},
		"ActiveState::Config::Tokenizer");
	    $child->{tokenizer} = $o->{tokenizer};
	}
    }
}

sub type {
    my $o = shift;
    my $type = $o->{type};
    if (@_) {
	my $t = $o->{type} = shift;
	delete $o->{children} unless $t eq 'group';
	delete $o->{value}    unless $t eq 'option';
	if ($t eq 'option' and $type eq 'group') {
	    $o->{name} = delete $o->{groupclass};
	}
	$o->dirty(1);
    }
    $type;
}

sub is_list {
    my $o = shift;
    return undef unless $o->{type} eq 'option';
    if (@_) {
	my $is_list = $o->is_list;
	my $list = shift;
	return $list if  $list and  defined $is_list;
	return undef if !$list and !defined $is_list;
	if ($list) {
	    my $v = $o->value; # I know it's a scalar
	    $o->{value} = defined $v ? [$v] : [];
	    $o->{line_comment} =
		defined $o->{line_comment} ? [$o->{line_comment}] : [];
	    $o->dirty(1);
	}
	else {
	    die "can't downgrade from a list to a single value!\n";
	}
	return;
    }
    my $v = $o->value;
    return ref($v) eq 'ARRAY' ? scalar @$v : undef;
}

sub _format {
    my ($o, $value) = @_;
    my (undef, undef, $f, $ts) = $o->typedecl;
    if ($ts) {
	# Walk through the list of formatters. Each validator may pop off
	# additional arguments from the @ts array.
	my @ts = @$ts;
	while (@ts) {
	    my $t = shift @ts;
	    next unless $t =~ /^#/; # must be an argument to the previous type
	    my (undef, $f, undef, $fargs) = $o->{parser}->typedecl($t);
	    next unless $f; # allow null formatters

	    # Skip those formatters which can't handle undef values:
	    next unless $o->{type} eq 'group'
		or defined $value or $fargs->{onundef};

	    # If this validator can't eat arrays, flatten the value:
	    if (ref($value) eq 'ARRAY' and !$fargs->{wantarray}) {
		my @save_ts = @ts;
		for my $w (@$value) {
		    @ts = @save_ts; # last iteration will modify @ts
		    $w = $f->($o, $w);
		}
	    }
	    else {
		$value = $f->($o, $value);
	    }
	}
    }
    elsif ($f) {
	return $f->($o, $value);
    }
    $value;
}

sub _validate {
    my ($o, $value, $apnd, $field) = @_;
    my $ctx =	defined $value		? 1		: 0;
    my $orig =	defined $value		? $value	:
		defined $field		? $o->{$field}	:
		undef;
    my $val = ref($orig) eq 'CODE' ? $o->$orig() : $orig;

    my $catch = sub {
	my $o = shift;
	if ($@) {
	    # Because we call out to external validators, we must be very
	    # careful to remove the dumb "at foo.pm line 7,..." that Perl
	    # inserts. Grumble: can we disable that?
	    $@ =~ s/\s+\z//;
	    $@ =~ s/^validation failed: //;
	    $@ =~ s/\bat .*$//;
	    die("validation failed at '$o->{path}': $@\n");
	}
    };

    my (undef, $v, undef, $ts) = $o->typedecl;
    if ($ts) {
	# Walk through the list of validators. Each validator may pop off
	# additional arguments from the @ts array. It may also modify $val and
	# $append.
	my @ts = @$ts;
	while (@ts) {
	    my $t = shift @ts;
	    next unless $t =~ /^#/; # must be an argument to the previous type
	    my ($v, undef, $vargs) = $o->{parser}->typedecl($t);
	    next unless $v; # allow null validators

	    # Skip those validators which can't handle undef values, unless
	    # this is a group node:
	    next unless $o->{type} eq 'group'
		or defined $val or $vargs->{onundef};

	    # If this validator can't eat arrays, flatten the value:
	    if (ref($val) eq 'ARRAY' and !$vargs->{wantarray}) {
		my @save_ts = @ts;
		for my $w (@$val) {
		    @ts = @save_ts; # the last iteration modifies @ts
		    eval {
			$v->($o->{parser}, $o->{path}, $o, $w, $apnd, $ctx, \@ts);
		    };
		    $o->$catch();
		}
	    }
	    # Otherwise give it the original value:
	    else {
		eval {
		    $v->($o->{parser}, $o->{path}, $o, $val, $apnd, $ctx, \@ts);
		};
		$o->$catch();
	    }
	}
    }
    # For an anonymous type, just call the validator method as-is:
    elsif ($v) {
	eval {
	    $v->($o->{parser}, $o->{path}, $o, $val, $apnd, $ctx, []);
	};
	$o->$catch();
    }
    return unless $ctx and defined $val;
    return if ref($orig) eq 'CODE';

    if ($field) {
	$o->{$field} = $val;
    }
    elsif ($apnd) {
	$o->is_list(1);
	push @{$o->{value}}, $val;
	$o->dirty(1);
    }
    else {
	$o->{value} = $val;
	$o->{line_comment} = $o->{line_comment}[0]
	    if ref($o->{line_comment}) eq 'ARRAY' and not defined $o->is_list;
	$o->dirty(1);
    }
}

sub set_to_default {
    my $o = shift;
    delete $o->{value};
    delete $o->{line_comment};
    $o->dirty(1);
}

sub value {
    my $o = shift;
    if (@_) {
	my $new = shift;
	$o->_validate($new, 0);
	$o->line_comment('') unless $o->line_comment;
	return;
    }
    my $val = exists $o->{value} ? $o->{value} : $o->{default};
    # For coderefs, we'll evaluate them every time this sub is called, until
    # they're overwritten with value($new_value). Because they may return
    # different values each time, the validator is called each time. In
    # practice, this means you should be careful when calling methods on the
    # node object from inside a validator, as you might create a recursive
    # loop.
    if (ref($val) eq 'CODE') {
	my $v1 = $o->$val();
	# save the validated value to $o->{__temp__}, then delete it.
	$o->_validate($v1, undef, '__temp__');
	$val = delete $o->{__temp__};
    }
    $o->changed(1);
    return $val unless ref($val) eq 'ARRAY';
    wantarray ? @$val : $val;
}

sub values {
    my $o = shift;
    return my @list_context = $o->value unless @_;
    $o->value([@_]);
    $o->line_comment('');
}

sub value_append {
    my $o = shift;
    my $new = shift;
    $o->_validate($new, 1);
    $o->line_comment('');
}

# Great utility method for getting the formatted value of the node without
# first fetching the value.
sub value_formatted {
    my $o = shift;
    return $o->_format($o->value);
}

sub children {
    my $o = shift;
    return undef unless $o->{type} eq 'group';
    my $children = $o->{children} ||= [];
    if (@_) {
	$o->{children} = $_[0]->all; # an iterator
	$o->dirty(1);
    }
    wantarray ? @$children : ActiveState::Config->iter($children);
}

sub block_comment {
    my $o = shift;
    my $com = $o->{block_comment};
    if (@_) {
	$o->{block_comment} = shift;
	$o->dirty(1);
    }
    $com;
}

sub line_comment {
    my $o = shift;
    my $is_list = $o->is_list;
    my $com = $o->{line_comment} ||= (defined $is_list ? [] : '');
    if (defined $is_list) {
	$o->{line_comment}[$_] = ''
	    for grep { !defined $o->{line_comment}[$_] } 0 .. $is_list - 1;
	if (@_) {
	    $o->{line_comment}[$is_list - 1] = shift;
	    $o->dirty(1);
	}
    }
    elsif (@_) {
	$o->{line_comment} = shift;
	$o->dirty(1);
    }
    return $com unless $is_list;
    wantarray ? @$com : $com;
}

sub typedecl {
    my $o = shift;
    my ($t, $v, $f, $ts) = @$o{qw(typedecl validator formatter typespec)};
    return wantarray ? ($t, $v, $f, $ts) : $t if $t;
    return wantarray ? () : undef;
}

sub _typedecl {
    my $o = shift;
    if (@_ >= 2 and ref($_[0]) eq 'CODE' and ref($_[1]) eq 'CODE') {
	@$o{qw(validator formatter)} = @_;
    }
    else {
	$o->{typedecl} = shift;
	my $nodefaults = shift;
	my @tspec = split ' ', $o->{typedecl};
	# Add the default type, unless told not to:
	if ($o->{parser}{opts}{default_type} and not $nodefaults) {
	    my ($t, @e) = @{$o->{parser}{opts}{default_type}};
	    my @t = split ' ', $t;
	    unshift @tspec, @t unless @e and grep { $tspec[0] eq $_ } @e;
	}
	$o->{typespec} = \@tspec;
    }
    $o->_validate(undef, undef, 'value');
}

sub _default {
    my ($o, $value) = @_;
    $o->{default} = $value;
    # If $o->{value} has been defined to some lame value like [] by the
    # validator, remove it.
    if (defined $o->{value}) {
	if (ref($o->{value}) eq 'ARRAY' and @{$o->{value}} == 0) {
	    delete $o->{value};
	}
    }
    $o->_validate(undef, undef, 'default');
}

sub tokenizer {
    my $o = shift;
    my $tok = $o->{tokenizer};
    $o->{tokenizer} = shift if @_;
    $tok;
}

sub path {
    my $o = shift;
    my $path = $o->{path};
    if (@_) {
	$o->{path} = shift;
	$o->{segment} = (split /\./, $o->{path})[-1];
	$o->dirty(1);
    }
    $path;
}

sub segment {
    my $o = shift;
    return $o->{segment};
}

sub parser {
    my $o = shift;
    $o->{parser};
}

# Return a native Perl structure suitable for dumping into a data tree.
sub _native {
    my ($o, $root) = @_;
    if ($o->{type} eq 'group') {
	my $tag = $o->{groupclass};
	my $name = $o->{name};
	my $children;
	if ($name) {
	    $children = $root->{$tag}{$name} ||= {};
	}
	else {
	    $children = $root->{$tag} ||= {};
	}

	# Check for any default values we should set, from the parser:
	my %map = $o->{parser}->_defaults_in_group($o->path);
	for my $path (keys %map) {
	    my $last = (split /\./, $path)[-1];
	    $children->{$last} = $map{$path};
	}

	# Do the real children
	for my $child ($o->children) {
	    $child->_native($children);
	}
    }
    elsif ($o->{name} !~ /^!/) {
	$root->{ $o->{name} } = $o->value;
    }
}

sub attach {
    my ($o, $path, $node) = @_;
    return undef unless $o->{type} eq 'group';
    $o->dirty(1);
    ActiveState::Config::_attach(@_);
}

sub _as_string {
    my $o = shift;
    my $emit = ActiveState::Config::Emitter->new($o, $o->{parser});
    my $txt = $emit->serialize;
}

sub _saveto {
    my ($o, $file) = @_;
    open my $FILE, "> $file" or die "can't write to $file: $!";
    print $FILE $o->_as_string;
    close $FILE or die "can't close $file: $!";
}

sub _save {
    my $o = shift;
    my $is_utf8 = $o->{parser}{opts}{is_utf8};
    ActiveState::Config::Emitter->new($o, $o->{parser}, $is_utf8)->save;
}

package ActiveState::Config::Iter;

sub I () { 0 }
sub V () { 1 }

sub cur {
    my $o = shift;
    $o->[V][$o->[I]];
}

sub next {
    my $o = shift;
    ++$o->[I];
    return if $o->[I] >= $o->size;
    $o->cur;
}

sub rewind {
    my $o = shift;
    $o->[I] = -1;
}

sub insert {
    my ($o, $node) = @_;
    $o->next if $o->[I] < 0;
    $node->dirty(1);
    splice @{$o->[V]}, $o->[I], 0, $node;
    --$o->[I];
}

sub append {
    my ($o, $node) = @_;
    $node->dirty(1);
    push @{$o->[V]}, $node;
}

sub delete {
    my $o = shift;
    my ($node) = splice @{$o->[V]}, $o->[I], 1;
    --$o->[I];
    $node->dirty(1);
    $node;
}

sub size {
    my $o = shift;
    scalar @{$o->[V]};
}

sub _all {
    my $o = shift;
    @{$o->[V]};
}

package ActiveState::Config::Types;
use ActiveState::Bytes qw(bytes_format bytes_parse);

sub register_types {
    my $o = shift;
    $o->typedecl('#bool',	\&v_bool,	\&f_bool);
    $o->typedecl('#int',	\&v_int,	\&f_int);
    $o->typedecl('#num',	\&v_num,	undef);
    $o->typedecl('#percentage',	\&v_per,	\&f_per);
    $o->typedecl('#duration',	\&v_dur,	\&f_dur);
    $o->typedecl('#enum',	\&v_enum,	undef);
    $o->typedecl('#str',	undef,		undef);
    $o->typedecl('#atom',	\&v_atom,	undef);
    $o->typedecl('#readonly',	\&v_ro,		undef);
    $o->typedecl('#size',       \&v_size,   \&f_size);
    $o->typedecl('#list',	\&v_list,	undef,	{
	wantarray	=> 1,
	onundef		=> 1,
    });
    $o->typedecl('#any',	undef,		undef);
}

sub VALUE_IX  () { 3 }
sub APPEND_IX () { 4 }
sub CTX_IX    () { 5 }

{
    my %bool = (
	yesno		=> ['no', 'yes'],
	onoff		=> ['off', 'on'],
	truefalse	=> ['false', 'true'],
	onezero		=> [0, 1],
    );
    sub v_bool {
	my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
	my $work = lc $v;

	# Perl likes to use "" as false by default; switch it to '0' so it
	# validates against the "onezero" method.
	$work = 0 if $v eq '';

	# Find what style they're using: true/false, yes/no, on/off, 1/0, etc.
	# Save the style they used, so that we can emit it that way later.
	for my $type (keys %bool) {
	    my ($val, $equiv);
	    for $equiv (0 .. $#{$bool{$type}}) {
		next unless $work eq $bool{$type}[$equiv];
		$val = $equiv;
		last;
	    }
	    next unless defined $val;

	    # Save the style used:
	    if ($type eq 'onezero') {
		$node->{__type__}{bool} ||= $type;
	    }
	    else {
		$node->{__type__}{bool} = $type;
	    }
	    $_[VALUE_IX] = $val;
	    $_[CTX_IX] = 1;
	    return 1;
	}
	die "invalid boolean value '$v'";
    }
    sub f_bool {
	my ($node, $val) = @_;
	my $ix = $val ? 1 : 0;
	return $bool{$node->{__type__}{bool} || 'truefalse'}[$ix];
    }
}

my $dec_rx = qr/([-+]?)([0-9]+)/;
my $oct_rx = qr/([-+]?)(0[0-7]+)/;
my $hex_rx = qr/([-+]?)(0[Xx][0-9A-Fa-f]+)/;

sub v_int {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    my $mode;
    if ($v =~ /^$oct_rx$/) {
	$mode = 'oct';
	my ($sign, $mantissa) = ($1, $2);
	$v = ($sign eq '-' ? -1 : 1) * oct($mantissa);
    }
    elsif ($v =~ /^$dec_rx$/) {
	$mode = 'dec';
	my ($sign, $mantissa) = ($1, $2);
	$v = ($sign eq '-' ? -1 : 1) * $mantissa;
    }
    elsif ($v =~ /^$hex_rx$/) {
	$mode = 'hex';
	my ($sign, $mantissa) = ($1, $2);
	$v = ($sign eq '-' ? -1 : 1) * hex($mantissa);
    }
    else { die "invalid integer '$v'" }
    $node->{__type__}{int} = $mode;
    $_[VALUE_IX] = $v;
    $_[CTX_IX] = 1;
    1;
}
sub f_int {
    my ($node, $val) = @_;
    my $neg = '';
    if ($val < 0) { $val = abs($val); $neg = '-' }
    if ($node->{__type__}{int} eq 'hex') {
	return sprintf("${neg}0x%x", $val);
    }
    elsif ($node->{__type__}{int} eq 'oct') {
	return sprintf("${neg}0%o", $val);
    }
    else {
	return "$neg$val";
    }
}

my $float_rx = qr/(?:(?i)(?:[+-]?)(?:(?=[0123456789]|[.])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[E])(?:(?:[+-]?)(?:[0123456789]+))|))/;

sub v_num {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    die "invalid number '$v'" unless $v =~ /^$float_rx$/;
    1;
}

sub v_per {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    die "invalid percentage '$v'" unless $v =~ s/^($float_rx)(\%?)$/$1/;
    $node->{__type__}{percentage} = $2;
    if ($2) {
	$_[VALUE_IX] = $v / 100;
	$_[CTX_IX] = 1;
    }
    1;
}
sub f_per {
    my ($node, $val) = @_;
    if ($node->{__type__}{percentage} or $val >= 0 and $val <= 1) {
	$val *= 100;
	return "$val%";
    }
    $val;
}

sub v_dur {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    my $tot = 0;
    my @parts = grep { length $_ } split m/($float_rx(?:)[smhdwy]?)/, $v;
    for my $part (@parts) {
	my ($n, $t) = ($part =~ /($float_rx)([smhdwy]?)/);
	die "invalid duration '$v'" unless defined $n;
	if    ($t eq 's' or !$t) { $tot += $n }
	elsif ($t eq 'm')        { $tot += $n * 60 }
	elsif ($t eq 'h')        { $tot += $n * 3600 }
	elsif ($t eq 'd')        { $tot += $n * 86400 }
	elsif ($t eq 'w')        { $tot += $n * 604800 }
	elsif ($t eq 'y')        { $tot += $n * 31536000 }
	else {
	    die "invalid duration '$v'";
	}
    }
    $node->{__type__}{duration} = $v;
    $_[VALUE_IX] = $tot;
    $_[CTX_IX] = 1;
    1;
}
sub f_dur {
    my ($node, $val) = @_;
    return "0s" unless $val;
    my $neg = $val =~ s/^-// ? '-' : '';
    my %t = map { $_ => 0 } qw(y w d h m s);
    my %v = (
	y => 31536000,
	w => 604800,
	d => 86400,
	h => 3600,
	m => 60,
    );
    for my $fac (qw(y w d h m)) {
	my $v = $val;
	while ($v >= 0) {
	    $v -= $v{$fac};
	    $t{$fac}++;
	}
	$t{$fac}--;
	$val = $v + $v{$fac};
	last unless $val;
    }
    $t{s} = $val;
    return $neg . join '', map { "$t{$_}$_" } grep { $t{$_} } qw(y w d h m s);
}

sub v_enum {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;

    # In an enumerated type, the allowed types are stored in @$ts:
    # $p->typedecl("foo.bar.baz", "#enum foo bar baz");
    # Therefore v_enum() does not forward anything to sub-types.
    my $matched = grep { $v eq $_ } @$ts;
    die "invalid value '$v': should be one of '@$ts'" unless $matched;
    @$ts = (); # consume the rest of the typespec
    1;
}

sub v_list {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;

    # Force list context on the node, and be sure the value does the
    # right thing.
    $node->type('option');
    if (!$append) {
	$_[CTX_IX]   = 1;
	if (defined $v) {
	    if (ref($v) eq 'ARRAY') {
		$_[VALUE_IX] = $v;
	    }
	    else {
		$_[VALUE_IX] = [$v];
		$node->{__emit__}{listtype} = 'repeated';
	    }
	}
	else {
	    $_[VALUE_IX] = [];
	}
    }
    1;
}

sub v_atom {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    my $mode = 'over';
    #$node->type('option');
    my $seen = $node->{__type__}{atom}{seen}++;
    $mode = shift @$ts if grep { @$ts and $ts->[0] eq $_ } qw(over nodups);
    die "node assigned twice"
	if $mode eq 'nodups' and $seen and $append and $ctx;
    die "node is a list, but is declared an atom"
	if ref($node->{value}) eq 'ARRAY';
    $_[APPEND_IX] = 0;
    1;
}

sub v_ro {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;

    # Refuse to let anybody set the value at all. Only the default value
    # is allowed.
    die "can't set readonly node" if $ctx;
}

sub v_size {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    die "Invalid size for $path '$v'\n" unless $v =~ /^(\d+)\s*([KMGTPEZYkmgtpezy]+|bytes)/;
    $_[3] = bytes_parse($v);
    $_[5] = 1;
    $_[3];
}

sub f_size {
    my ($node, $val) = @_;
    $val = bytes_format($val);
    $val;
}

1;
